"""Creates the Windfarm class/model."""
from __future__ import annotations

import csv
import logging
import datetime as dt
import itertools
from math import fsum

import numpy as np
import pandas as pd
import networkx as nx
from geopy import distance

from wombat.core import RepairManager, WombatEnvironment
from wombat.core.library import load_yaml
from wombat.windfarm.system import Cable, System
from wombat.core.data_classes import String, SubString, WindFarmMap, SubstationMap
from wombat.utilities.utilities import cache


class Windfarm:
    """The primary class for operating on objects within a windfarm. The substations,
    cables, and turbines are created as a network object to be more appropriately
    accessed and controlled.
    """

    def __init__(
        self,
        env: WombatEnvironment,
        windfarm_layout: str,
        repair_manager: RepairManager,
    ) -> None:
        self.env = env
        self.repair_manager = repair_manager

        # Set up the layout and instantiate all windfarm objects
        self._create_graph_layout(windfarm_layout)
        self._create_turbines_and_substations()
        self._create_cables()
        self.capacity: int | float = sum(
            self.system(turb).capacity for turb in self.turbine_id
        )
        self._create_substation_turbine_map()
        self._create_wind_farm_map()
        self.finish_setup()
        self.calculate_distance_matrix()

        # Create the logging items
        self.system_list = list(self.graph.nodes)
        self._setup_logger()

        # Register the windfarm and start the logger
        self.repair_manager._register_windfarm(self)
        self.env._register_windfarm(self)
        self.env.process(self._log_operations())

    def _create_graph_layout(self, windfarm_layout: str) -> None:
        """Creates a network layout of the windfarm start from the substation(s) to
        be able to capture downstream turbines that can be cut off in the event of a
        cable failure.

        Parameters
        ----------
        windfarm_layout : str
            Filename to use for reading in the windfarm layout; must be a csv file.
        """
        try:
            layout_path = str(self.env.data_dir / "project/plant" / windfarm_layout)
            layout = (
                pd.read_csv(layout_path)
                .sort_values(by=["string", "order"])
                .reset_index(drop=True)
            )
        except FileNotFoundError:
            layout_path = str(self.env.data_dir / "windfarm" / windfarm_layout)
            layout = (
                pd.read_csv(layout_path)
                .sort_values(by=["string", "order"])
                .reset_index(drop=True)
            )
            logging.warning(
                "DeprecationWarning: In v0.8, all wind farm layout files must be"
                " located in: '<library>/project/plant/"
            )
        layout.subassembly = layout.subassembly.fillna("")
        layout.upstream_cable = layout.upstream_cable.fillna("")

        windfarm = nx.DiGraph()
        windfarm.add_nodes_from(layout.id.values)

        # Assign the data attributes to the graph nodes
        for col in ("name", "latitude", "longitude", "subassembly"):
            nx.set_node_attributes(windfarm, dict(layout[["id", col]].values), name=col)

        # Determine which nodes are substations and which are turbines
        if "type" in layout.columns:
            if layout.loc[~layout.type.isin(("substation", "turbine"))].size > 0:
                raise ValueError(
                    "At least one value in the 'type' column are not one of:"
                    " 'substation' or 'turbine'."
                )
            substation_filter = layout.type == "substation"
            nx.set_node_attributes(
                windfarm, dict(layout[["id", "type"]].values), name="type"
            )
        else:
            substation_filter = layout.id == layout.substation_id
            _type = {True: "substation", False: "turbine"}
            d = {i: _type[val] for i, val in zip(layout.id, substation_filter.values)}
            nx.set_node_attributes(windfarm, d, name="type")

        self.substation_id = layout.loc[substation_filter, "id"].values
        for substation in self.substation_id:
            windfarm.nodes[substation]["connection"] = layout.loc[
                layout.id == substation, "substation_id"
            ].values[0]

        self.turbine_id = layout.loc[~substation_filter, "id"].values
        substations = layout[substation_filter].copy()
        turbines = layout[~substation_filter].copy()
        substation_sections = [
            turbines[turbines.substation_id == substation]
            for substation in substations.id
        ]
        for section in substation_sections:
            for _, row in section.iterrows():
                if row.order == 0:
                    start: str = row.substation_id
                else:
                    start = current  # noqa: F821
                current: str = row.id
                windfarm.add_edge(
                    start, current, length=row.distance, cable=row.upstream_cable
                )
        for substation in self.substation_id:
            row = layout.loc[layout.id == substation]
            windfarm.add_edge(
                row.substation_id.values[0],
                substation,
                length=row.distance.values[0],
                cable=row.upstream_cable.values[0],
            )

        self.graph = windfarm
        self.layout_df = layout

    def _create_turbines_and_substations(self) -> None:
        """Instantiates the turbine and substation models as defined in the
        user-provided layout file, and connects these models to the appropriate graph
        nodes to create a fully representative windfarm network model.

        Raises
        ------
        ValueError
            Raised if the subassembly data is not provided in the layout file.
        """
        bad_data_location_messages = []
        for system_id, data in self.graph.nodes(data=True):
            if data["subassembly"] == "":
                raise ValueError(
                    "A 'subassembly' file must be specified for all nodes in the"
                    " windfarm layout!"
                )

            try:
                subassembly_dict = load_yaml(
                    self.env.data_dir / f"{data['type']}s", data["subassembly"]
                )
            except FileNotFoundError:
                subassembly_dict = load_yaml(
                    self.env.data_dir / "windfarm", data["subassembly"]
                )
                message = (
                    f"In v0.7, all {data['type']} configurations must be located in:"
                    f" '<library>/{data['type']}s"
                )
                bad_data_location_messages.append(message)
            self.graph.nodes[system_id]["system"] = System(
                self.env,
                self.repair_manager,
                system_id,
                data["name"],
                subassembly_dict,
                data["type"],
            )

        # Raise the warning for soon-to-be deprecated library structure
        bad_data_location_messages = list(set(bad_data_location_messages))
        for message in bad_data_location_messages:
            logging.warning(f"DeprecationWarning: {message}")

    def _create_cables(self) -> None:
        """Instantiates the cable models as defined in the user-provided layout file,
        and connects these models to the appropriate graph edges to create a fully
        representative windfarm network model.

        Raises
        ------
        ValueError
            Raised if the cable model is not specified.
        """
        get_name = "upstream_cable_name" in self.layout_df
        bad_data_location_messages = []
        for start_node, end_node, data in self.graph.edges(data=True):
            # Check that the cable data is provided
            if data["cable"] == "":
                raise ValueError(
                    "A 'cable' file must be specified for all nodes in the"
                    " windfarm layout!"
                )
            try:
                cable_dict = load_yaml(self.env.data_dir / "cables", data["cable"])
            except FileNotFoundError:
                cable_dict = load_yaml(self.env.data_dir / "windfarm", data["cable"])
                bad_data_location_messages.append(
                    "In v0.7, all cable configurations must be located in:"
                    " '<library>/cables/"
                )

            start_coordinates = (
                self.graph.nodes[start_node]["latitude"],
                self.graph.nodes[start_node]["longitude"],
            )
            end_coordinates = (
                self.graph.nodes[end_node]["latitude"],
                self.graph.nodes[end_node]["longitude"],
            )

            name = None
            if get_name:
                name, *_ = self.layout_df.loc[
                    self.layout_df.id == end_node, "upstream_cable_name"
                ]

            # If the real distance/cable length is not input, then the geodesic distance
            # is calculated
            if data["length"] == 0:
                data["length"] = distance.geodesic(
                    start_coordinates, end_coordinates, ellipsoid="WGS-84"
                ).km

            if self.graph.nodes[end_node]["type"] == "substation":
                data["type"] = "export"
            else:
                data["type"] = "array"

            data["cable"] = Cable(
                self, self.env, data["type"], start_node, end_node, cable_dict, name
            )

            # Calaculate the geometric center point
            end_points = np.array((start_coordinates, end_coordinates))
            data["latitude"], data["longitude"] = end_points.mean(axis=0)

        # Raise the warning for soon-to-be deprecated library structure
        bad_data_location_messages = list(set(bad_data_location_messages))
        for message in bad_data_location_messages:
            logging.warning(f"DeprecationWarning: {message}")

    def calculate_distance_matrix(self) -> None:
        """Calculates the geodesic distance, in km, between all of the windfarm's nodes,
        e.g., substations and turbines, and cables.
        """
        ids = list(self.graph.nodes())
        ids.extend([data["cable"].id for *_, data in self.graph.edges(data=True)])
        coords = [
            (data["latitude"], data["longitude"])
            for *_, data in (*self.graph.nodes(data=True), *self.graph.edges(data=True))
        ]

        dist = [
            distance.geodesic(c1, c2).km for c1, c2 in itertools.combinations(coords, 2)
        ]
        dist_arr = np.ones((len(ids), len(ids)))
        triangle_ix = np.triu_indices_from(dist_arr, 1)
        dist_arr[triangle_ix] = dist_arr.T[triangle_ix] = dist

        # Set the self distance to infinity, so that only one crew can be dropped off
        # at a single point
        np.fill_diagonal(dist_arr, np.inf)

        self.distance_matrix = pd.DataFrame(dist_arr, index=ids, columns=ids)

    def _create_substation_turbine_map(self) -> None:
        """Creates ``substation_turbine_map``, a dictionary, that maps substation(s) to
        the dependent turbines in the windfarm, and the weighting of each turbine in the
        windfarm.
        """
        # Get all turbines connected to each substation, excepting any connected via
        # export cables that connect substations as these operate independently
        s_t_map: dict = {s: {"turbines": [], "weights": []} for s in self.substation_id}
        for substation_id in self.substation_id:
            nodes = set(
                nx.bfs_tree(self.graph, substation_id, depth_limit=1).nodes
            ).difference(self.substation_id)
            for node in list(nodes):
                nodes.update(
                    list(itertools.chain(*nx.dfs_successors(self.graph, node).values()))
                )
            s_t_map[substation_id]["turbines"] = np.array(list(nodes))

        # Reorient the mapping to have the turbine list and the capacity-based weighting
        # of each turbine
        for s_id in s_t_map:
            s_t_map[s_id]["weights"] = (
                np.array([self.system(t).capacity for t in s_t_map[s_id]["turbines"]])
                / self.capacity
            )

        self.substation_turbine_map: dict[str, dict[str, np.ndarray]] = s_t_map

    def _create_wind_farm_map(self) -> None:
        """Creates a secondary graph object strictly for traversing the windfarm to turn
        on/off the appropriate turbines, substations, and cables more easily.
        """
        substations = self.substation_id
        graph = self.graph
        wind_map = dict(zip(substations, itertools.repeat({})))

        export = [el for el in graph.edges if el[1] in substations]
        for cable_tuple in export:
            self.cable(cable_tuple).set_string_details(*cable_tuple[::-1])

        for s_id in self.substation_id:
            start_nodes = list(
                set(nx.bfs_tree(graph, s_id, depth_limit=1).nodes).difference(
                    substations
                )
            )
            wind_map[s_id] = {"strings": dict(zip(start_nodes, itertools.repeat({})))}

            for start_node in start_nodes:
                upstream = list(
                    itertools.chain(*nx.dfs_successors(graph, start_node).values())
                )
                wind_map[s_id]["strings"][start_node] = {
                    start_node: SubString(
                        downstream=s_id,  # type: ignore
                        upstream=upstream,  # type: ignore
                    )
                }
                self.cable((s_id, start_node)).set_string_details(start_node, s_id)

                downstream = start_node
                for node in upstream:
                    wind_map[s_id]["strings"][start_node][node] = SubString(
                        downstream=downstream,  # type: ignore
                        upstream=list(  # tye: ignore
                            itertools.chain(*nx.dfs_successors(graph, node).values())
                        ),
                    )
                    self.cable((downstream, node)).set_string_details(start_node, s_id)
                    downstream = node

                wind_map[s_id]["strings"][start_node] = String(  # type: ignore
                    start=start_node, upstream_map=wind_map[s_id]["strings"][start_node]
                )
            wind_map[s_id] = SubstationMap(  # type: ignore
                string_starts=start_nodes,
                string_map=wind_map[s_id]["strings"],
                downstream=graph.nodes[s_id]["connection"],
            )
        self.wind_farm_map = WindFarmMap(
            substation_map=wind_map, export_cables=export  # type: ignore
        )

    def finish_setup(self) -> None:
        """Final initialization hook for any substations, turbines, or cables."""
        for start_node, end_node in self.graph.edges():
            self.cable((start_node, end_node)).finish_setup()

    def _setup_logger(self, initial: bool = True):
        self._log_columns = [
            "datetime",
            "env_datetime",
            "env_time",
        ] + self.system_list

        self.env._operations_writer = csv.DictWriter(
            self.env._operations_csv, delimiter="|", fieldnames=self._log_columns
        )
        if initial:
            self.env._operations_writer.writeheader()

    def _log_operations(self):
        """Logs the operational data for a simulation."""
        message = {
            "datetime": dt.datetime.now(),
            "env_datetime": self.env.simulation_time,
            "env_time": self.env.now,
        }
        message.update(
            {system: self.system(system).operating_level for system in self.system_list}
        )
        self.env._operations_writer.writerow(message)

        HOURS = 1
        while True:
            for _ in range(10000):
                yield self.env.timeout(HOURS)
                message = {
                    "datetime": dt.datetime.now(),
                    "env_datetime": self.env.simulation_time,
                    "env_time": self.env.now,
                }
                message.update(
                    {
                        system: self.system(system).operating_level
                        for system in self.system_list
                    }
                )
                self.env._operations_buffer.append(message)
            self.env._operations_writer.writerows(self.env._operations_buffer)
            self.env._operations_buffer.clear()

    @cache
    def system(self, system_id: str) -> System:
        """Convenience function to returns the desired `System` object for a turbine or
        substation in the windfarm.

        Parameters
        ----------
        system_id : str
            The system's unique identifier, ``wombat.windfarm.System.id``.

        Returns
        -------
        System
            The ``System`` object.
        """
        return self.graph.nodes[system_id]["system"]

    @cache
    def cable(self, cable_id: tuple[str, str] | str) -> Cable:
        """Convenience function to returns the desired `Cable` object for a cable in the
        windfarm.

        Parameters
        ----------
        cable_id : tuple[str, str] | str
            The cable's unique identifier, of the form: (``wombat.windfarm.System.id``,
            ``wombat.windfarm.System.id``), for the (downstream node id, upstream node
            id), or the ``Cable.id``.

        Returns
        -------
        Cable
            The ``Cable`` object.
        """
        if isinstance(cable_id, str):
            edge_id = tuple(cable_id.split("::")[1:])
        else:
            edge_id = cable_id
        try:
            return self.graph.edges[edge_id]["cable"]
        except KeyError:
            raise KeyError(f"Edge {edge_id} is invalid.")

    @property
    def current_availability(self) -> float:
        r"""Calculates the product of all system ``operating_level`` variables across
        the windfarm using the following forumation.

        .. math::
            \sum{
                OperatingLevel_{substation_{i}} *
                \sum{OperatingLevel_{turbine_{j}} * Weight_{turbine_{j}}}
            }

        where the :math:``{OperatingLevel}`` is the product of the operating level
        of each subassembly on a given system (substation or turbine), and the
        :math:``{Weight}`` is the proportion of one turbine's capacity relative to
        the whole windfarm.
        """  # noqa: W605
        operating_levels = {
            s_id: [
                self.system(t).operating_level
                for t in self.substation_turbine_map[s_id]["turbines"]  # type: ignore
            ]
            for s_id in self.substation_turbine_map
        }
        availability = fsum(
            [
                self.system(s_id).operating_level
                * fsum(
                    operating_levels[s_id]
                    * self.substation_turbine_map[s_id]["weights"]  # type: ignore
                )
                for s_id in self.substation_turbine_map
            ]
        )
        return availability

    @property
    def current_availability_wo_servicing(self) -> float:
        r"""Calculates the product of all system ``operating_level`` variables across
        the windfarm using the following forumation, ignoring 0 operating level due to
        ongoing servicing.

        .. math::
            \sum{
                OperatingLevel_{substation_{i}} *
                \sum{OperatingLevel_{turbine_{j}} * Weight_{turbine_{j}}}
            }

        where the :math:``{OperatingLevel}`` is the product of the operating level
        of each subassembly on a given system (substation or turbine), and the
        :math:``{Weight}`` is the proportion of one turbine's capacity relative to
        the whole windfarm.
        """  # noqa: W605
        operating_levels = {
            s_id: [
                self.system(t).operating_level_wo_servicing
                for t in self.substation_turbine_map[s_id]["turbines"]  # type: ignore
            ]
            for s_id in self.substation_turbine_map
        }
        availability = fsum(
            [
                self.system(s_id).operating_level_wo_servicing
                * fsum(
                    operating_levels[s_id]
                    * self.substation_turbine_map[s_id]["weights"]  # type: ignore
                )
                for s_id in self.substation_turbine_map
            ]
        )
        return availability



"""Provides unit tests for the `System` class."""
import numpy as np
import pytest
import numpy.testing as npt

from wombat.core import RepairManager
from wombat.windfarm.system import System

from tests.conftest import (
    SUBSTATION,
    VESTAS_V90,
    VESTAS_POWER_CURVE,
    VESTAS_V90_1_SUBASSEMBLY,
    VESTAS_V90_NO_SUBASSEMBLY,
)


def test_turbine_initialization_complete_setup(env_setup):
    """Tests a complete turbine setup."""
    ENV = env_setup
    MANAGER = RepairManager(ENV)
    turbine_id = "WTG001"
    turbine_name = "Vestas V90 - 001"

    system = System(
        env=ENV,
        repair_manager=MANAGER,
        t_id=turbine_id,
        name=turbine_name,
        subassemblies=VESTAS_V90,
        system="turbine",
    )
    assert system.env == ENV
    assert system.repair_manager == MANAGER
    assert system.id == turbine_id
    assert system.name == turbine_name
    assert system.capacity == VESTAS_V90["capacity_kw"]
    assert system.value == VESTAS_V90["capacity_kw"] * VESTAS_V90["capex_kw"]
    subassemblies = [
        "electrical_system",
        "electronic_control",
        "sensors",
        "hydraulic_system",
        "yaw_system",
        "rotor_blades",
        "mechanical_brake",
        "rotor_hub",
        "gearbox",
        "generator",
        "supporting_structure",
        "drive_train",
    ]
    for subassembly in subassemblies:
        assert hasattr(system, subassembly)

    # Check that only the turbine subassemblies are created
    assert not hasattr(system, "transformer")

    windspeed = np.arange(0, 30, 0.2)
    correct_power = VESTAS_POWER_CURVE(windspeed)
    npt.assert_equal(system.power_curve(windspeed), correct_power)
    npt.assert_equal(system.power(windspeed), correct_power)

    # Test that we're starting at 100% operations
    assert system.operating_level == 1.0


def test_turbine_initialization_minimal_setup(env_setup):
    """Test an incomplete, or partial turbine defintion for correctness."""
    ENV = env_setup
    MANAGER = RepairManager(ENV)
    turbine_id = "WTG001"
    turbine_name = "Vestas V90 - 001"

    system = System(
        env=ENV,
        repair_manager=MANAGER,
        t_id=turbine_id,
        name=turbine_name,
        subassemblies=VESTAS_V90_1_SUBASSEMBLY,
        system="turbine",
    )
    assert system.env == ENV
    assert system.repair_manager == MANAGER
    assert system.id == turbine_id
    assert system.name == turbine_name
    assert system.capacity == VESTAS_V90["capacity_kw"]
    assert system.value == VESTAS_V90["capacity_kw"] * VESTAS_V90["capex_kw"]

    # Test that only the generator subassembly was created
    missing_subassemblies = [
        "electrical_system",
        "electronic_control",
        "sensors",
        "hydraulic_system",
        "yaw_system",
        "rotor_blades",
        "mechanical_brake",
        "rotor_hub",
        "gearbox",
        "supporting_structure",
        "drive_train",
    ]
    for subassembly in missing_subassemblies:
        assert not hasattr(system, subassembly)

    assert hasattr(system, "generator")

    # Check that only the turbine subassemblies are created
    assert not hasattr(system, "transformer")

    # Check that there is no power curve defined, aside from all zero output
    windspeed = np.arange(0, 30, 0.2)
    correct_power = np.zeros(windspeed.shape)
    npt.assert_equal(system.power_curve(windspeed), correct_power)
    npt.assert_equal(system.power(windspeed), correct_power)

    # Test that we're starting at 100% operations
    assert system.operating_level == 1.0


def test_substation_initialization(env_setup):
    """Tests a complete substation setup."""
    ENV = env_setup
    MANAGER = RepairManager(ENV)
    substation_id = "OSS001"
    substation_name = "SUBSTATION - 001"

    system = System(
        env=ENV,
        repair_manager=MANAGER,
        t_id=substation_id,
        name=substation_name,
        subassemblies=SUBSTATION,
        system="substation",
    )
    assert system.env == ENV
    assert system.repair_manager == MANAGER
    assert system.id == substation_id
    assert system.name == substation_name
    assert system.capacity == SUBSTATION["capacity_kw"]
    assert system.value == SUBSTATION["capacity_kw"] * SUBSTATION["capex_kw"]

    # Check that only the turbine subassemblies are created
    assert hasattr(system, "transformer")

    # Check that none of the turbine subassemblies are defined
    subassemblies = [
        "electrical_system",
        "electronic_control",
        "sensors",
        "hydraulic_system",
        "yaw_system",
        "rotor_blades",
        "mechanical_brake",
        "rotor_hub",
        "gearbox",
        "generator",
        "supporting_structure",
        "drive_train",
    ]
    for subassembly in subassemblies:
        assert not hasattr(system, subassembly)

    # Test that we're starting at 100% operations
    assert system.operating_level == 1.0


def test_initialization_error_cases(env_setup):
    """Tests the cases where a System definition should fail."""
    ENV = env_setup
    MANAGER = RepairManager(ENV)

    # No subassemblies provided
    turbine_id = "WTG001"
    turbine_name = "Vestas V90 - 001"
    with pytest.raises(ValueError):
        System(
            env=ENV,
            repair_manager=MANAGER,
            t_id=turbine_id,
            name=turbine_name,
            subassemblies=VESTAS_V90_NO_SUBASSEMBLY,
            system="turbine",
        )

    # Incorrect system definitions
    incorrect_system = "invalid"
    with pytest.raises(ValueError):
        System(
            env=ENV,
            repair_manager=MANAGER,
            t_id=turbine_id,
            name=turbine_name,
            subassemblies=VESTAS_V90,
            system=incorrect_system,
        )

    incorrect_system = "turbbine"
    with pytest.raises(ValueError):
        System(
            env=ENV,
            repair_manager=MANAGER,
            t_id=turbine_id,
            name=turbine_name,
            subassemblies=VESTAS_V90,
            system=incorrect_system,
        )

    incorrect_system = "ssubstation"
    with pytest.raises(ValueError):
        System(
            env=ENV,
            repair_manager=MANAGER,
            t_id=turbine_id,
            name=turbine_name,
            subassemblies=VESTAS_V90,
            system=incorrect_system,
        )


def test_operating_level(env_setup):
    """Tests the ``operating_level`` method."""
    ENV = env_setup
    MANAGER = RepairManager(ENV)
    turbine_id = "WTG001"
    turbine_name = "Vestas V90 - 001"

    system = System(
        env=ENV,
        repair_manager=MANAGER,
        t_id=turbine_id,
        name=turbine_name,
        subassemblies=VESTAS_V90,
        system="turbine",
    )

    # Test the starting operating level
    assert system.operating_level == 1

    # Test updating a single subassembly
    system.generator.operating_level = 0.9
    assert system.operating_level == 0.9

    # Test updating another subassembly
    system.electronic_control.operating_level = 0.9
    assert system.operating_level == 0.9 * 0.9

    # Test a cable failure
    system.cable_failure = ENV.event()
    assert system.operating_level == 0.0
    system.cable_failure.succeed()

    # Test a failed subassembly shuts down the turbine
    system.generator.operating_level = 0.0
    assert system.operating_level == 0.0



"""Tests the Subassembly class."""


import numpy as np
import pytest

from wombat.windfarm import Windfarm
from wombat.windfarm.system import System, Subassembly
from wombat.core.data_classes import Failure, Maintenance
from wombat.core.repair_management import RepairManager

from tests.conftest import (
    SUBSTATION,
    VESTAS_V90,
    GENERATOR_SUBASSEMBLY,
    VESTAS_V90_TEST_TIMEOUTS,
)


np.random.seed(2022)  # for test_interruptions()


def test_subassembly_initialization(env_setup):
    """Test the initialization of a subassembly."""
    # Define the basic items
    ENV = env_setup
    MANAGER = RepairManager(ENV)
    TURBINE = System(ENV, MANAGER, "WTG001", "Vestas V90 001", VESTAS_V90, "turbine")
    OSS = System(ENV, MANAGER, "OSS1", "OSS1", SUBSTATION, "substation")
    TRANSFORMER = SUBSTATION["transformer"]

    # Test the initialization for a turbine and generator
    subassembly_id = "GEN001"
    data_dict = GENERATOR_SUBASSEMBLY  # same data as the one in the library
    generator = Subassembly(TURBINE, ENV, subassembly_id, data_dict)
    correct_N_maintenance = len(GENERATOR_SUBASSEMBLY["maintenance"])
    correct_N_failures = len(GENERATOR_SUBASSEMBLY["failures"])

    assert generator.env == ENV
    assert generator.system == TURBINE
    assert generator.id == subassembly_id
    assert generator.name == data_dict["name"]
    # assert subassembly.data == TURBINE.generator.data  # TODO: equality testing
    assert generator.broken.triggered
    assert generator.operating_level == 1.0
    assert len(generator.processes) == correct_N_maintenance + correct_N_failures
    # Maintenance keys are m{i} and failure keys are integer severity
    N_failures = len([el for el in generator.processes if isinstance(el, int)])
    assert N_failures == correct_N_failures

    # Run the same tests for a substation and transformer
    subassembly_id = "TRNS1"
    data_dict = TRANSFORMER
    correct_N_maintenance = len(TRANSFORMER["maintenance"])
    correct_N_failures = len(TRANSFORMER["failures"])
    transformer = Subassembly(OSS, ENV, subassembly_id, data_dict)

    assert transformer.env == ENV
    assert transformer.system == OSS
    assert transformer.id == subassembly_id
    assert transformer.name == data_dict["name"]
    # assert subassembly.data == OSS.transformer.data  # TODO: equality testing
    assert transformer.broken.triggered
    assert transformer.operating_level == 1.0
    assert len(transformer.processes) == correct_N_maintenance + correct_N_failures
    # Maintenance keys are m{i} and failure keys are integer severity
    N_failures = len([el for el in transformer.processes if isinstance(el, int)])
    assert N_failures == correct_N_failures


@pytest.mark.skip(reason="The timing of the failures needs to be updated")
def test_interruptions_and_request_submission(env_setup):
    """Test the initialization of a subassembly.

    NOTE: This test will not run correctly if not running the whole pytest suite because
    it relies on the when/where the the random variable is set in the process and the
    order in which the files run. This is considered to be acceptable behavior because
    it is not testing for the actual timing, but the behavior when certain events happen
    and that they are logged.
    """
    # Define the basic items
    ENV = env_setup
    MANAGER = RepairManager(ENV)
    TURBINE = System(
        ENV, MANAGER, "WTG001", "Vestas V90 001", VESTAS_V90_TEST_TIMEOUTS, "turbine"
    )

    # Timeout turbine will have (randomness controlled by random seed at the top):
    #  1) a generator maintenance event every 5 days (120 hours/time steps)
    #  2) a generator failure event with a Weibull of scale 0.5, shape 1, with
    #      frequencies at:
    #      2.47, 13.47, 303.92 hours (last one not reached)
    #  3) a gearbox failure event with a Weibull of scale 5, shape 1, with
    #      frequencies at:
    #      1054.20 and 9132.90 hours (last one not reached)
    #  4) a generator catastrophic failure event at 3027.74 hours that stops everything

    # The resets and maintenance tasks will be removed from the record for the
    # generator upon failure
    correct_N_generator_failures = 2  # 2 resets reached prior to catastrophic failure
    correct_N_generator_maintenance = 4  # 4 prior to failure
    correct_N_gearbox_failures = 1  # none reached
    correct_N_gearbox_maintenance = 0  # none reached

    # PROCESS FOR DETERMINING ACTUAL EVENTS IN CASE OF CHANGE OF RANDOM BEHAVIOR
    """
    # TODO
    t = 2.47 + 13.47 + 303.92 + 175.41 + 599.03 + 597.26
    print(t)
    ENV.run(1 + t + 2)
    print("generator")
    for key, process in TURBINE.generator.processes.items():
        print(key, process.__dict__)
    print("gearbox")
    for key, process in TURBINE.gearbox.processes.items():
        print(key, process.__dict__)
    print(ENV.now)
    assert False
    """

    catastrophic_timeout = 527.1  # see notes above for catastrophic failure timing

    # Test that all the failures that will get purged actually happen
    ENV.run(catastrophic_timeout - 1)

    requests = MANAGER.items
    generator_requests = [el for el in requests if el.subassembly_id == "generator"]
    gearbox_requests = [el for el in requests if el.subassembly_id == "gearbox"]
    generator_maintenance = [
        el for el in generator_requests if isinstance(el.details, Maintenance)
    ]
    generator_failures = [
        el for el in generator_requests if isinstance(el.details, Failure)
    ]
    gearbox_maintenance = [
        el for el in gearbox_requests if isinstance(el.details, Maintenance)
    ]
    gearbox_failures = [
        el for el in gearbox_requests if isinstance(el.details, Failure)
    ]

    # See the notes above for timing
    assert len(generator_failures) == correct_N_generator_failures
    assert len(generator_maintenance) == correct_N_generator_maintenance
    assert len(gearbox_failures) == correct_N_gearbox_failures
    assert len(gearbox_maintenance) == correct_N_gearbox_maintenance

    # Test that the correct failures/tasks get purged and all the processes are
    # timed out appropriately
    ENV.run(catastrophic_timeout + 1)

    # Check that the generator is set to broken
    assert TURBINE.generator.broken is True
    assert TURBINE.gearbox.broken is False
    assert TURBINE.generator.operating_level == 0
    assert TURBINE.gearbox.operating_level == 1
    assert TURBINE.operating_level == 0

    # Check that the 24 hour default timeout has been setup, which indicates that all
    # processes have been interrupted. This is the 1 day timeout to see if the turbine
    # is operating again in case of repairs
    for process in TURBINE.generator.processes.values():
        assert process._target._delay == 24
    for process in TURBINE.gearbox.processes.values():
        assert process._target._delay == 24

    requests = MANAGER.items
    generator_requests = [el for el in requests if el.subassembly_id == "generator"]
    generator_maintenance = [
        el for el in generator_requests if isinstance(el.details, Maintenance)
    ]
    generator_failures = [
        el for el in generator_requests if isinstance(el.details, Failure)
    ]
    gearbox_requests = [el for el in requests if el.subassembly_id == "gearbox"]
    gearbox_maintenance = [
        el for el in gearbox_requests if isinstance(el.details, Maintenance)
    ]
    gearbox_failures = [
        el for el in gearbox_requests if isinstance(el.details, Failure)
    ]

    # Check the total number of request submitted from each subassembly is correct
    assert len(generator_requests) == 1  # all others will have been purged
    assert (
        len(gearbox_requests)
        == correct_N_gearbox_failures + correct_N_gearbox_maintenance
    )

    # Check the number of maintenance and failure requests for each subassembly
    assert len(gearbox_failures) == correct_N_gearbox_failures
    assert len(gearbox_maintenance) == correct_N_gearbox_maintenance
    assert len(generator_failures) == 1  # all but catastrophic failure removed
    assert len(generator_maintenance) == 0  # all removed

    # Run for another 24 hours to ensure nothing has changed
    ENV.run(catastrophic_timeout + 24)

    assert TURBINE.generator.broken is True
    assert TURBINE.gearbox.broken is False
    assert TURBINE.generator.operating_level == 0
    assert TURBINE.gearbox.operating_level == 1
    assert TURBINE.operating_level == 0

    for process in TURBINE.generator.processes.values():
        assert process._target._delay == 24
    for process in TURBINE.gearbox.processes.values():
        assert process._target._delay == 24

    # Run until the end to ensure nothig has changed
    ENV.run(ENV.max_run_time)

    assert TURBINE.generator.broken is True
    assert TURBINE.gearbox.broken is False
    assert TURBINE.generator.operating_level == 0
    assert TURBINE.gearbox.operating_level == 1
    assert TURBINE.operating_level == 0

    for process in TURBINE.generator.processes.values():
        assert process._target._delay == 24
    for process in TURBINE.gearbox.processes.values():
        assert process._target._delay == 24


def test_timeouts_for_zeroed_out(env_setup):
    """Tests that the timeouts for any zeroed out subassembly maintenance or failure
    does not occur.
    """
    # Define the basic items
    ENV = env_setup
    MANAGER = RepairManager(ENV)
    TURBINE = System(ENV, MANAGER, "WTG001", "Vestas V90 001", VESTAS_V90, "turbine")
    Windfarm(ENV, "layout.csv", MANAGER)

    # Run the simulation 1 timestep to get it started, then inspec the process timeouts
    ENV.run(1)

    # Only the generator should have timeouts not equal to max run time
    for subassembly in TURBINE.subassemblies:
        if subassembly.id == "generator":
            continue
        for process in subassembly.processes.values():
            assert process._target._delay == ENV.max_run_time

    # Catastrophic failure at 284508.82985483576
    # TODO: Don't have a way to test that all update to max_run_time - failure time



"""The servicing equipment module provides a small number of utility functions specific
to the operations of servicing equipment and the `ServiceEquipment` class that provides
the repair and transportation logic for scheduled, unscheduled, and unscheduled towing
servicing equipment.
"""
# TODO: NEED A SPECIFIC STARTUP METHOD
from __future__ import annotations

import logging
from copy import deepcopy
from math import ceil
from typing import TYPE_CHECKING, Any, Generator
from pathlib import Path
from datetime import timedelta
from itertools import zip_longest

import numpy as np
import pandas as pd
from simpy.events import Event, Process, Timeout
from pandas.core.indexes.datetimes import DatetimeIndex

from wombat.core import (
    Maintenance,
    RepairManager,
    RepairRequest,
    WombatEnvironment,
    ServiceEquipmentData,
)
from wombat.windfarm import Windfarm
from wombat.utilities import HOURS_IN_DAY, cache, hours_until_future_hour
from wombat.core.mixins import RepairsMixin
from wombat.core.library import load_yaml
from wombat.windfarm.system import System
from wombat.core.data_classes import (
    UNSCHEDULED_STRATEGIES,
    ScheduledServiceEquipmentData,
    UnscheduledServiceEquipmentData,
)
from wombat.windfarm.system.cable import Cable
from wombat.windfarm.system.subassembly import Subassembly


if TYPE_CHECKING:
    from wombat.core import Port


# Numpy random generation initialization
random_generator = np.random.default_rng(seed=42)


def consecutive_groups(data: np.ndarray, step_size: int = 1) -> list[np.ndarray]:
    """Generates the subgroups of an array where the difference between two sequential
    elements is equal to the ``step_size``. The intent is to find the length of delays
    in a weather forecast.

    Parameters
    ----------
    data : np.ndarray
        An array of integers.
    step_size : int, optional
        The step size to be considered a consecutive number, by default 1.

    Returns
    -------
    list[np.ndarray]
        A list of arrays of the consecutive elements of ``data``.
    """
    # Consecutive groups in delay
    groups = np.split(data, np.where(np.diff(data) != step_size)[0] + 1)
    return groups


def calculate_delay_from_forecast(
    forecast: np.ndarray, hours_required: int | float
) -> tuple[bool, int]:
    """Calculates the delay from the binary weather forecast for if that hour is all
    clear for operations.

    Parameters
    ----------
    forecast : np.ndarray
        Truth array to indicate if that hour satisfies the weather limit requirements.
    hours_required : np.ndarray
        The minimum clear weather window required, in hours.

    Returns
    -------
    tuple[bool, int]
        Indicator if a window is found (``True``) or not (``False``), and the number
        of hours the event needs to be delayed in order to start.
    """
    safe_operating_windows = consecutive_groups(np.where(forecast)[0])
    window_lengths = np.array([window.size for window in safe_operating_windows])
    clear_windows = np.where(window_lengths >= hours_required)[0]
    if clear_windows.size == 0:
        return False, forecast.size
    return True, safe_operating_windows[clear_windows[0]][0]


def validate_end_points(start: str, end: str, no_intrasite: bool = False) -> None:
    """Checks the starting and ending locations for traveling and towing.

    Parameters
    ----------
    start : str
        The starting location; should be on of: "site", "system", or "port".
    end : str
        The ending location; should be on of: "site", "system", or "port".
    no_intrasite : bool
        A flag to disable intrasite travel, so that ``start`` and ``end`` cannot
        both be "system", by default False.

    Raises
    ------
    ValueError
        Raised if the starting location is invalid.
    ValueError
        Raised if the ending location is invalid
    ValueError
        Raised if "system" is provided to both ``start`` and ``end``, but
        ``no_intrasite`` is set to ``True``.
    """
    if start not in ("port", "site", "system"):
        raise ValueError(
            "``start`` location must be one of 'port', 'site', or 'system'!"
        )
    if end not in ("port", "site", "system"):
        raise ValueError("``end`` location must be one of 'port', 'site', or 'system'!")
    if no_intrasite and (start == end == "system"):
        raise ValueError("No travel within the site is allowed for this process")


def reset_system_operations(system: System) -> None:
    """Completely resets the failure and maintenance events for a given system
    and its subassemblies, and puts each ``Subassembly.operating_level`` back to 100%.

    .. note:: This is only intended to be used in conjunction with a tow-to-port
        repair where a turbine will be completely serviced.

    Parameters
    ----------
    system : System
        The turbine to be reset.
    """
    for subassembly in system.subassemblies:
        subassembly.operating_level = 1.0
        subassembly.recreate_processes()


class ServiceEquipment(RepairsMixin):
    """Provides a servicing equipment object that can handle various maintenance and
    repair tasks.

    Parameters
    ----------
    env : WombatEnvironment
        The simulation environment.
    windfarm : Windfarm
        The ``Windfarm`` object.
    repair_manager : RepairManager
        The ``RepairManager`` object.
    equipment_data_file : str
        The equipment settings file name with extension.

    Attributes
    ----------
    env : WombatEnvironment
        The simulation environment instance.
    windfarm : Windfarm
        The simulation windfarm instance.
    manager : RepairManager
        The simulation repair manager instance.
    settings : ScheduledServiceEquipmentData | UnscheduledServiceEquipmentData
        The servicing equipment's configuration settings, as provided by the user.
    onsite : bool
        Indicates if the servicing equipment is at the site (``True``), or not
        (``False``).
    enroute : bool
        Indicates if the servicing equipment is on its way to the site (``True``),
        or not (``False``).
    at_port : bool
        Indicates if the servicing equipment is at the port, or similar location for
        land-based, (``True``), or not (``False``).
    at_system : bool
        Indications if the servicing equipment is at a cable, substation, or turbine
        while on the site (``True``), or not (``False``).
    current_system : str | None
        Either the ``System.id`` if ``at_system``, or ``None`` if not.
    transferring_crew : bool
        Indications if the servicing equipment is at a cable, substation, or turbine
        and transferring the crew to or from that system (``True``), or not
        (``False``).
    """

    def __init__(
        self,
        env: WombatEnvironment,
        windfarm: Windfarm,
        repair_manager: RepairManager,
        equipment_data_file: str | Path | dict,
    ):
        """Initializes the ``ServiceEquipment`` class.

        Parameters
        ----------
        env : WombatEnvironment
            The simulation environment.
        windfarm : Windfarm
            The ``Windfarm`` object.
        repair_manager : RepairManager
            The ``RepairManager`` object.
        equipment_data_file : str
            The equipment settings file name with extension.
        """
        self.env = env
        self.windfarm = windfarm
        self.manager = repair_manager
        self.onsite = False  # True: mobilized to the site, but not necessarily at_site
        self.dispatched = False
        self.enroute = False
        self.at_port = False
        self.at_site = False
        self.at_system = False
        self.transferring_crew = False
        self.current_system = None  # type: str | None
        self.port_based = False  # Changed to False if port is registered
        self.settings: ScheduledServiceEquipmentData | UnscheduledServiceEquipmentData

        if isinstance(equipment_data_file, (str, Path)):
            try:
                data = load_yaml(env.data_dir / "vessels", equipment_data_file)
            except FileNotFoundError:
                data = load_yaml(env.data_dir / "repair/transport", equipment_data_file)
                logging.warning(
                    "DeprecationWarning: In v0.8, all servicing equipment"
                    " configurations must be located in: '<library>/vessels/"
                )
        else:
            data = equipment_data_file

        try:
            if data["start_year"] < self.env.start_year:
                data["start_year"] = self.env.start_year
            if data["end_year"] > self.env.end_year:
                data["end_year"] = self.env.end_year
        except KeyError:
            # Ignores for unscheduled maintenace equipment that won't have this input
            pass

        # NOTE: mypy is not caught up with attrs yet :(
        self.settings = ServiceEquipmentData(data).determine_type()  # type: ignore

        # Register servicing equipment with the repair manager if it is using an
        # unscheduled maintenance scenario, so it can be dispatched as needed
        if self.settings.strategy in UNSCHEDULED_STRATEGIES:
            self.manager._register_equipment(self)

        # Only run the equipment if it is on a scheduled basis, otherwise wait
        # for it to be dispatched
        if self.settings.strategy == "scheduled":
            self.env.process(self.run_scheduled_in_situ())

        # Create partial functions for the labor and equipment costs for clarity
        self.initialize_cost_calculators(which="equipment")

    def finish_setup_with_environment_variables(self) -> None:
        """A post-initialization step that will override unset parameters with those
        from the the environemt that may have already been set.
        """
        self._check_working_hours(which="env")
        self.settings._set_port_distance(self.env.port_distance)
        self.settings.set_non_operational_dates(
            self.env.non_operational_start,
            self.env.start_year,
            self.env.non_operational_end,
            self.env.end_year,
        )
        self.settings.set_reduced_speed_parameters(
            self.env.reduced_speed_start,
            self.env.start_year,
            self.env.reduced_speed_end,
            self.env.end_year,
            self.env.reduced_speed,
        )

    def _register_port(self, port: Port) -> None:
        """Method for a tugboat at attach the port for two-way communications. This also
        sets the vessel to be at the port, and updates the port_distance.

        Parameters
        ----------
        port : Port
            The port where the tugboat is based.
        """
        self.port = port
        self.port_based = True
        self.at_port = True
        self.settings._set_port_distance(port.settings.site_distance)  # type: ignore

        self._check_working_hours(which="port")

        # Set the non-operational start/end dates if needed
        self.settings.set_non_operational_dates(
            port.settings.non_operational_start,
            self.env.start_year,
            port.settings.non_operational_end,
            self.env.end_year,
        )

        # Set the reduced speed start/end dates if needed
        self.settings.set_reduced_speed_parameters(
            port.settings.reduced_speed_start,
            self.env.start_year,
            port.settings.reduced_speed_end,
            self.env.end_year,
            port.settings.reduced_speed,
        )

    def _set_location(self, end: str, set_current: str | None = None) -> None:
        """Keeps track of the servicing equipment by setting the location at either:
        site, port, or a specific system.

        Parameters
        ----------
        end : str
            The ending location; one of "site", or "port"
        set_current : str
            The ``System.id`` for the new current location, if one is to be set.
        """
        if end == "port":
            self.at_port = True
            self.at_site = False
        else:
            self.at_port = False
            self.at_site = True
        self.at_system = True if set_current is not None else False
        self.current_system = set_current

    def _weather_forecast(
        self, hours: int | float, which: str
    ) -> tuple[DatetimeIndex, np.ndarray, np.ndarray]:
        """Retrieves the weather forecast from the simulation environment, and
        translates it to a boolean for satisfactory (True) and unsatisfactory (False)
        weather conditions.

        Parameters
        ----------
        hours : int | float
            The number of hours of weather data that should be retrieved.
        which : str
            One of "repair" or "transport" to indicate which weather limits to be using.

        Returns
        -------
        tuple[DatetimeIndex, np.ndarray, np.ndarray]
            The pandas DatetimeIndex, the hour of day array, and the boolean array of
            where the weather conditions are within safe operating limits for the
            servicing equipment (True) or not (False).
        """
        if which == "repair":
            max_wind = self.settings.max_windspeed_repair
            max_wave = self.settings.max_waveheight_repair
        elif which == "transport":
            max_wind = self.settings.max_windspeed_transport
            max_wave = self.settings.max_waveheight_transport
        else:
            raise ValueError("`which` must be one of `repair` or `transport`.")

        dt, hour, wind, wave = self.env.weather_forecast(hours)
        all_clear = (wind <= max_wind) & (wave <= max_wave)
        return dt, hour, all_clear

    def get_speed(self, tow: bool = False) -> float:
        """Determines the appropriate speed that the servicing equipment should be
        traveling at for towing or traveling, and if the timeframe is during a reduced
        speed scenario.

        Parameters
        ----------
        tow : bool, optional
            True indicates the servicing equipment should be using the towing speed,
            and if False, then the traveling speed should be used, by default False.

        Returns
        -------
        float
            The maximum speed the servicing equipment should be traveling/towing at.
        """
        assert hasattr(self.settings, "reduced_speed_dates_set")
        speed = self.settings.tow_speed if tow else self.settings.speed  # type: ignore
        if self.env.simulation_time.date() in self.settings.reduced_speed_dates_set:
            if speed > self.settings.reduced_speed:
                speed = self.settings.reduced_speed
        return speed

    def get_next_request(self):
        """Gets the next request by the rig's method for processing repairs.

        Returns
        -------
        simpy.resources.store.FilterStoreGet
            The next ``RepairRequest`` to be processed.
        """
        if self.settings.method == "turbine":
            return self.manager.get_request_by_system(self.settings.capability)
        if self.settings.method == "severity":
            return self.manager.get_request_by_severity(self.settings.capability)

    def enable_string_operations(self, cable: Cable) -> None:
        """Traverses the upstream cable and turbine connections and resets the
        ``System.cable_failure`` and ``Cable.downstream_failure`` until it hits
        another cable failure, then the loop exits.

        Parameters
        ----------
        subassembly : Cable
            The `Cable` or `System`
        """
        farm = self.manager.windfarm
        if cable.connection_type == "array":
            # If there is another failure downstream of the repaired cable, do nothing
            if not cable.downstream_failure.triggered:
                return

            # For each upstream turbine and cable, reset their operations
            nodes = deepcopy(cable.upstream_nodes)
            cables = cable.upstream_cables
            tid = nodes.pop(0)
            turbine = farm.system(tid)
            turbine.cable_failure.succeed()
            for tid, cid in zip_longest(nodes, cables, fillvalue=None):  # type: ignore
                if cid is not None:
                    cable = farm.cable(cid)
                    cable.downstream_failure.succeed()
                    if not cable.broken.triggered:
                        break
                if tid is not None:  # only None for last cable on string
                    turbine = farm.system(tid)
                    turbine.cable_failure.succeed()

        if cable.connection_type == "export":
            # Reset the substation's cable failure
            farm.system(cable.end_node).cable_failure.succeed()

            # For each string connected to the substation reset all the turbines and
            # cables until another cable failure is encoountered, then move to the next
            # string
            for t_list, c_list in zip(cable.upstream_nodes, cable.upstream_cables):
                for t, c in zip_longest(t_list, c_list, fillvalue=None):  # type: ignore
                    if c is not None:
                        cable = farm.cable(c)
                        if not cable.broken.triggered:
                            break
                        cable.downstream_failure.succeed()
                    farm.system(t).cable_failure.succeed()

    def register_repair_with_subassembly(
        self,
        subassembly: Subassembly | Cable,
        repair: RepairRequest,
        starting_operating_level: float,
    ) -> None:
        """Goes into the repaired subassembly, component, or cable and returns its
        ``operating_level`` back to good as new for the specific repair. For fatal cable
        failures, all upstream turbines are turned back on unless there is another fatal
        cable failure preventing any more from operating.

        Parameters
        ----------
        subassembly : Subassembly | Cable
            The subassembly or cable that was repaired.
        repair : RepairRequest
            The request for repair that was submitted.
        starting_operating_level : float
            The operating level before a repair was started.
        """
        operation_reduction = repair.details.operation_reduction

        # Put the subassembly/component back to good as new condition
        if repair.details.replacement:
            subassembly.operating_level = 1.0
            _ = self.manager.purge_subassembly_requests(
                repair.system_id, repair.subassembly_id
            )
        if operation_reduction == 1:
            subassembly.operating_level = starting_operating_level
            subassembly.broken.succeed()
        elif operation_reduction == 0:
            subassembly.operating_level = starting_operating_level
        else:
            subassembly.operating_level /= 1 - operation_reduction

        if isinstance(subassembly, Subassembly):
            self.manager.enable_requests_for_system(subassembly.system)
        elif isinstance(subassembly, Cable):
            # If the system is a cable, re-enable the upstream systems and cables
            self.manager.enable_requests_for_system(subassembly)
            if operation_reduction == 1:
                self.enable_string_operations(subassembly)
        else:
            raise TypeError("`subassembly` was neither a `Cable` nor `Subassembly`.")

        self.env.process(self.manager.register_repair(repair))

    def wait_until_next_operational_period(
        self, *, less_mobilization_hours: int = 0
    ) -> Generator[Timeout, None, None]:
        """Delays the crew and equipment until the start of the next operational
        period.

        TODO: Need a custom error if weather doesn't align with the equipment dates.

        Parameters
        ----------
        less_mobilization_hours : int
            The number of hours required for mobilization that will be subtracted from
            the waiting period to account for mobilization, by default 0.

        Yields
        ------
        Generator[Timeout, None, None]
            A Timeout event for the number of hours between when the function is called
            and when the next operational period starts.
        """
        assert isinstance(self.settings, ScheduledServiceEquipmentData)
        current = self.env.simulation_time.date()
        ix_match = np.where(current < self.settings.operating_dates)[0]
        if ix_match.size > 0:
            still_in_operation = True
            next_operating_date = self.settings.operating_dates[ix_match[0]]
            hours_to_next_shift = (
                self.env.date_ix(next_operating_date)
                - self.env.now
                + self.settings.workday_start
                - less_mobilization_hours
            )
        else:
            still_in_operation = False
            hours_to_next_shift = self.env.max_run_time - self.env.now

        if still_in_operation:
            # TODO: This message needs to account for unscheduled equipment as well, but
            # the post processor filtering needs to be updated as well
            additional = "will return next year"
        else:
            additional = "no more return visits will be made"

        # Ensures that the statuses are correct
        self.at_port = False
        self.at_site = False
        self.enroute = False
        self.onsite = False

        self.env.log_action(
            agent=self.settings.name,
            action="delay",
            reason="work is complete",
            additional=additional,
            duration=hours_to_next_shift,
            location="enroute",
        )

        yield self.env.timeout(hours_to_next_shift)

    def mobilize_scheduled(self) -> Generator[Timeout, None, None]:
        """Mobilizes the ServiceEquipment object by waiting for the next operational
        period, less the mobilization time, then logging the mobiliztion cost.

        NOTE: weather delays are not accounted for in this process.

        Yields
        ------
        Generator[Timeout, None, None]
            A Timeout event for the number of hours between when the function is called
            and when the next operational period starts.
        """
        mobilization_hours = self.settings.mobilization_days * HOURS_IN_DAY
        yield self.env.process(  # type: ignore
            self.wait_until_next_operational_period(
                less_mobilization_hours=mobilization_hours
            )
        )
        self.env.log_action(
            agent=self.settings.name,
            action="mobilization",
            reason=f"{self.settings.name} is being mobilized",
            additional="mobilization",
            location="enroute",
        )

        self.enroute = True
        yield self.env.timeout(mobilization_hours)
        self.onsite = True
        self.enroute = False

        self.env.log_action(
            agent=self.settings.name,
            action="mobilization",
            reason=f"{self.settings.name} has arrived on site",
            additional="mobilization",
            duration=mobilization_hours,
            equipment_cost=self.settings.mobilization_cost,
            location="site",
        )

    def mobilize(self) -> Generator[Timeout, None, None]:
        """Mobilizes the ServiceEquipment object.

        NOTE: weather delays are not accounted for at this stage.

        Yields
        ------
        Generator[Timeout, None, None]
            A Timeout event for the number of hours the ServiceEquipment requires for
            mobilizing to the windfarm site.
        """
        self.env.log_action(
            agent=self.settings.name,
            action="mobilization",
            reason=f"{self.settings.name} is being mobilized",
            additional="mobilization",
            location="enroute",
        )

        self.enroute = True
        mobilization_hours = self.settings.mobilization_days * HOURS_IN_DAY
        yield self.env.timeout(mobilization_hours)
        self.onsite = True
        self.enroute = False

        self.env.log_action(
            agent=self.settings.name,
            action="mobilization",
            reason=f"{self.settings.name} has arrived on site",
            additional="mobilization",
            duration=mobilization_hours,
            equipment_cost=self.settings.mobilization_cost,
            location="site",
        )

    def find_uninterrupted_weather_window(
        self, hours_required: int | float
    ) -> tuple[int | float, bool]:
        """Finds the delay required before starting on a process that won't be able to
        be interrupted by a weather delay.

        TODO: WEATHER FORECAST NEEDS TO BE DONE FOR ``math.floor(self.now)``, not the
        ceiling or there will be a whole lot of rounding up errors on process times.

        Parameters
        ----------
        hours_required : int | float
            The number of uninterrupted of hours that a process requires for completion.

        Returns
        -------
        tuple[int | float, bool]
            The number of hours in weather delays before a process can be completed, and
            an indicator for if the process has to be delayed until the next shift for
            a safe transfer.
        """
        # If the hours required for the window is 0, then return 0 and indicate there is
        # no shift delay to be processed
        if hours_required == 0:
            return 0, False

        current = self.env.simulation_time

        # If the time required for a transfer is longer than the time left in the shift,
        # then return the hours left in the shift and indicate a shift delay
        max_hours = hours_until_future_hour(
            self.env.simulation_time, self.settings.workday_end
        )
        if hours_required > max_hours:
            return max_hours, True

        _, hour, all_clear = self._weather_forecast(max_hours, which="repair")
        all_clear &= self._is_workshift(hour)
        safe_operating_windows = consecutive_groups(np.where(all_clear)[0])
        window_lengths = np.array([window.size for window in safe_operating_windows])

        # If all the safe operating windows are shorter than the time required, then
        # return the time until the end of the shift and indicate a shift delay
        if all([window < hours_required for window in window_lengths]):
            return max_hours, True

        # Find the length of the delay
        delay = safe_operating_windows[
            np.where(window_lengths >= hours_required)[0][0]
        ][0]

        # If there is no delay, simply return 0 and the shift delay indicator
        if delay == 0:
            return delay, False

        # If the delay is non-zero, ensure we get the correct hour difference from now
        # until the top of the first available hour
        delay = hours_until_future_hour(current, current.hour + delay)
        return delay, False

    def find_interrupted_weather_window(
        self, hours_required: int | float
    ) -> tuple[DatetimeIndex, np.ndarray, bool]:
        """Assesses at which points in the repair window, the wind (and wave)
        constraints for safe operation are met.

        The initial search looks for a weather window of length ``hours_required``, and
        adds 24 hours to the window for the proceeding 9 days. If no satisfactory window
        is found, then the calling process must make another call to this function, but
        likely there is something wrong with the constraints or weather conditions if
        no window is found within 10 days.

        Parameters
        ----------
        hours_required : int | float
            The number of hours required to complete the repair.

        Returns
        -------
        tuple[DatetimeIndex, np.ndarray, bool]
            The pandas DatetimeIndex, and a corresponding boolean array for what points
            in the time window are safe to complete a maintenance task or repair.
        """
        window_found = False
        hours_required = ceil(hours_required)

        for i in range(10):  # no more than 10 attempts to find a window
            dt_ix, hour_ix, weather = self._weather_forecast(
                hours_required + (i * 24), which="repair"
            )
            working_hours = self._is_workshift(hour_ix)
            window = weather & working_hours
            if window.sum() >= hours_required:
                window_found = True
                break

        return dt_ix, weather, window_found

    def weather_delay(self, hours: int | float, **kwargs) -> Generator[Event, Any, Any]:
        """Processes a weather delay of length ``hours`` hours. If ``hours`` = 0, then
        a Timeout is still processed, but not logging is done (this is to increase
        consistency and do better typing validation across the codebase).

        Parameters
        ----------
        hours : int | float
            The lenght, in hours, of the weather delay.

        Yields
        ------
        Generator[Event, Any, Any]
            If the delay is more than 0 hours, then a ``Timeout`` is yielded of length
            ``hours``.
        """
        if hours < 0:
            raise ValueError(
                f"`hours` must be greater than 0 for {self.settings.name} to process"
                " a weather delay"
            )
        if hours == 0:
            return

        salary_cost = self.calculate_salary_cost(hours)
        hourly_cost = 0  # contractors not paid for delays
        equipment_cost = self.calculate_equipment_cost(hours)
        self.env.log_action(
            duration=hours,
            action="delay",
            additional="weather delay",
            salary_labor_cost=salary_cost,
            hourly_labor_cost=hourly_cost,
            equipment_cost=equipment_cost,
            **kwargs,
        )
        yield self.env.timeout(hours)

    @cache
    def _calculate_intra_site_time(
        self, start: str | None, end: str | None
    ) -> tuple[float, float]:
        """Calculates the time it takes to travel between port and site or between
        systems on site.

        Parameters
        ----------
        start : str | None
            The starting onsite location. If ``None``, then 0 is returned.
        end : str | None
            The ending onsite location. If ``None``, then 0 is returned.

        Returns
        -------
        tuple[float, float]
            The travel time and distance between two locations.
        """
        distance = 0.0  # setting for invalid cases to have no traveling
        valid_sys = self.windfarm.distance_matrix.columns
        intra_site = start in valid_sys and end in valid_sys
        if intra_site:
            distance = self.windfarm.distance_matrix.loc[start, end]

        # if no speed is set, then there is no traveling time
        speed = self.get_speed()
        travel_time = distance / speed if speed > 0 else 0.0

        # Infinity is the result of "traveling" between a system twice in a row
        if travel_time == float("inf"):
            travel_time = 0
        return travel_time, distance

    def _calculate_uninterrupted_travel_time(
        self, distance: float, tow: bool = False
    ) -> tuple[float, float]:
        """Calculates the delay to the start of traveling and the amount of time it
        will take to travel between two locations.

        Parameters
        ----------
        distance : float
            The distance to be traveled.
        tow : bool
            Indicates if this travel is for towing (True), or not (False), by default
            False.

        Returns
        -------
        tuple[float, float]
            The length of the delay and the length of travel time, in hours.` -1 is
            returned if there no weather windows, and the process will have to be
            attempted again.
        """
        # Return 0 delay and travel time in the distance is zero
        if distance == 0:
            return 0, 0

        speed = self.get_speed(tow=tow)
        hours_required = distance / speed

        n = 1
        max_extra_days = 4
        while n <= max_extra_days:  # max tries before failing
            hours = hours_required + (24 * n)
            *_, all_clear = self._weather_forecast(hours, which="repair")
            found, delay = calculate_delay_from_forecast(all_clear, hours_required)
            if found:
                return delay, hours_required
            n += 1

        # Return -1 for delay if no weather window was found
        return -1, hours_required

    def _calculate_interrupted_travel_time(
        self, distance: float, tow: bool = False
    ) -> float:
        """Calculates the travel time with speed reductions for inclement weather, but
        without shift interruptions.

        Parameters
        ----------
        distance : flaot
            The total distance to be traveled, in km.
        tow : bool
            Indicates if this travel is for towing (True), or not (False), by default
            False.

        Returns
        -------
        float
            _description_
        """
        speed = self.get_speed(tow=tow)
        reduction_factor = 1 - self.settings.speed_reduction_factor
        reduction_factor = 0.01 if reduction_factor == 0 else reduction_factor

        # get the weather forecast with this time for the max travel time
        max_hours = 1 + distance / speed * (1 / reduction_factor)
        dt, _, all_clear = self._weather_forecast(max_hours, which="transport")

        # calculate the distance able to be traveled in each 1-hour window
        distance_traveled = speed * all_clear.astype(float)
        distance_traveled[distance_traveled == 0] = speed * reduction_factor

        # Reduce the first time step by the time lapsed since the start of the hour
        # before proceeding
        distance_traveled[0] *= 1 - self.env.now % 1

        # Cumulative sum at the end of each full hour
        distance_traveled_sum = distance_traveled.cumsum()

        # Get the index for the end of the hour where the distance requred to be
        # traveled is reached.
        try:
            ix_hours = np.where(distance_traveled_sum >= distance)[0][0]
        except IndexError as e:
            # If an error occurs because an index maxes out the weather window, check
            # that it's not due to having reached the end of the simulation. If so,
            # return the max amount of time, but if that's not the case re-raise the
            # error.
            if self.env.weather.index.values[-1] in dt:
                ix_hours = distance_traveled.size - 1
            else:
                raise e

        # Shave off the extra timing to get the exact travel time
        total_hours = ix_hours + 1  # add 1 for 0-indexing
        traveled = distance_traveled_sum[ix_hours]
        if traveled > distance:
            difference = traveled - distance
            speed_at_hour = distance_traveled[ix_hours]
            reduction = difference / speed_at_hour
            total_hours -= reduction

        return total_hours

    def travel(
        self,
        start: str,
        end: str,
        set_current: str | None = None,
        hours: float | None = None,
        distance: float | None = None,
        **kwargs,
    ) -> Generator[Timeout | Process, None, None]:
        """The process for traveling between port and site, or two systems onsite.

        NOTE: This does not currently take the weather conditions into account.

        Parameters
        ----------
        start : str
            The starting location, one of "site", "port", or "system".
        end : str
            The starting location, one of "site", "port", or "system".
        set_current : str, optional
            Where to set ``current_system`` to be if traveling to site or a different
            system onsite, by default None.
        hours : float, optional
            The number hours required for traveling between ``start`` and ``end``.
            If provided, no internal travel time will be calculated.
        distance : float, optional
            The distance, in km, to be traveled. Only used if hours is provided

        Yields
        ------
        Generator[Timeout | Process, None, None]
            The timeout event for traveling.
        """
        validate_end_points(start, end)
        if hours is None:
            hours = 0.0
            distance = 0.0
            additional = f"traveling from {start} to {end}"

            if start == end == "system":
                additional = f"traveling from {self.current_system} to {set_current}"
                hours, distance = self._calculate_intra_site_time(
                    self.current_system, set_current
                )
            elif {start, end} == {"site", "port"}:
                additional = f"traveling from {start} to {end}"
                distance = self.settings.port_distance
                hours = self._calculate_interrupted_travel_time(distance)
        else:
            if distance is None:
                raise ValueError("`distance` must be provided if `hours` is provided.")

        # MyPy helpers
        assert isinstance(distance, (int, float))
        assert isinstance(hours, (float, int))

        # If the the equipment will arive after the shift is over, then it must travel
        # back to port (if needed), and wait for the next shift
        future_time = self.env.simulation_time + timedelta(hours=hours)
        is_shift = self._is_workshift(future_time.hour)
        if not is_shift and end != "port" and not self.at_port:
            kw = {
                "additional": "insufficient time to complete travel before end of the shift"  # noqa: disabl#501
            }
            kw.update(kwargs)
            yield self.env.process(
                self.travel(start=start, end="port", **kw)  # type: ignore
            )
            yield self.env.process(self.wait_until_next_shift(**kwargs))
            yield self.env.process(
                self.travel(start=start, end=end, set_current=set_current, **kwargs)
            )
            return
        elif not is_shift and self.at_port:
            kw = {
                "additional": "insufficient time to complete travel before end of the shift"  # noqa: disabl#501
            }
            kw.update(kwargs)
            yield self.env.process(self.wait_until_next_shift(**kw))
            yield self.env.process(
                self.travel(start=start, end=end, set_current=set_current, **kwargs)
            )
            return

        salary_cost = self.calculate_salary_cost(hours)
        hourly_cost = 0  # contractors not paid for traveling
        equipment_cost = self.calculate_equipment_cost(hours)
        kwargs.update({"additional": additional})
        self.env.log_action(
            action="traveling",
            duration=hours,
            distance_km=distance,
            salary_labor_cost=salary_cost,
            hourly_labor_cost=hourly_cost,
            equipment_cost=equipment_cost,
            location="enroute",
            **kwargs,
        )

        # Unregister current_system during travel <- partial fix, still need to figure
        # out where current_system needs to be set to None to allow things to "work"
        self._set_location("site")

        yield self.env.timeout(hours)

        self._set_location(end, set_current)
        where = set_current if set_current is not None else end
        kwargs.update({"additional": f"arrived at {where}"})
        self.env.log_action(
            action="complete travel",
            location=end,
            **kwargs,
        )

    def tow(
        self,
        start: str,
        end: str,
        set_current: str | None = None,
        **kwargs,
    ) -> Generator[Timeout | Process, None, None]:
        """The process for towing a turbine to/from port.

        Parameters
        ----------
        start : str
            The starting location; one of "site" or "port".
        end : str
            The ending location; one of "site" or "port".
        set_current : str | None, optional
            The ``System.id`` if the turbine is being returned to site, by default None

        Yields
        ------
        Generator[Timeout | Process, None, None]
            The series of SimPy events that will be processed for the actions to occur.
        """
        validate_end_points(start, end, no_intrasite=True)
        # Get the distance that needs to be traveled, then calculate the delay and time
        # traveling, and log each of them
        distance = self.settings.port_distance
        delay, hours = self._calculate_uninterrupted_travel_time(distance, tow=True)
        if delay == -1:
            if start == "site":
                kw = deepcopy(kwargs)
                kw.update({"reason": "Insufficient weather window, return to port"})
                yield self.env.process(self.travel("site", "port", **kw))
                yield self.env.timeout(HOURS_IN_DAY * 4)
                yield self.env.process(
                    self.tow(start, end, set_current=set_current, **kwargs)
                )
            elif start == "port":
                kw = deepcopy(kwargs)
                kw.update(
                    {"reason": "Insufficient weather window, will try again later"}
                )
                yield self.env.timeout(HOURS_IN_DAY * 4)
                yield self.env.process(
                    self.tow(start, end, set_current=set_current, **kwargs)
                )

        else:
            self.env.process(self.weather_delay(delay, location=end, **kwargs))

        salary_labor_cost = self.calculate_salary_cost(hours)
        hourly_labor_cost = self.calculate_hourly_cost(hours)
        equipment_cost = self.calculate_equipment_cost(hours)
        self.env.log_action(
            duration=hours,
            distance_km=distance,
            action="towing",
            salary_labor_cost=salary_labor_cost,
            hourly_labor_cost=hourly_labor_cost,
            equipment_cost=equipment_cost,
            location="enroute",
            **kwargs,
        )

        yield self.env.timeout(hours)
        self._set_location(end, set_current)
        self.env.log_action(
            action="complete towing",
            salary_labor_cost=salary_labor_cost,
            hourly_labor_cost=hourly_labor_cost,
            equipment_cost=equipment_cost,
            additional="complete",
            location=end,
            **kwargs,
        )

    def crew_transfer(
        self,
        system: System | Cable,
        subassembly: Subassembly,
        request: RepairRequest,
        to_system: bool,
    ) -> Generator[Timeout | Process, None, None]:
        """The process of transfering the crew from the equipment to the ``System``
        for servicing using an uninterrupted weather window to ensure safe transfer.

        Parameters
        ----------
        system : System | Cable
            The System where the crew needs to be transferred to.
        subassembly : Subassembly
            The Subassembly that is being worked on.
        request : RepairRequest
            The repair to be processed.
        to_system : bool
            True if the crew is being transferred to the system, or False if the crew
            is being transferred off the system.

        Returns
        -------
        None
            None is returned when this is run recursively to not repeat the crew
            transfer process.

        Yields
        ------
        Generator[Timeout | Process, None, None]
            Yields a timeout event for the crew transfer once an uninterrupted weather
            window can be found.
        """
        hours_to_process = self.settings.crew_transfer_time
        salary_cost = self.calculate_salary_cost(hours_to_process)
        hourly_cost = self.calculate_hourly_cost(hours_to_process)
        equipment_cost = self.calculate_equipment_cost(hours_to_process)

        shared_logging = {
            "system_id": system.id,
            "system_name": system.name,
            "part_id": subassembly.id,
            "part_name": subassembly.name,
            "system_ol": system.operating_level,
            "part_ol": subassembly.operating_level,
            "agent": self.settings.name,
            "reason": request.details.description,
            "request_id": request.request_id,
        }

        delay, shift_delay = self.find_uninterrupted_weather_window(hours_to_process)
        # If there is a shift delay, then travel to port, wait, and travel back, and
        # finally try again.
        if shift_delay:
            travel_time = self.env.now
            yield self.env.process(
                self.travel(start="site", end="port", **shared_logging)
            )
            travel_time -= self.env.now  # will be negative value, but is flipped
            delay -= abs(travel_time)  # decrement the delay by the travel time

            # If removing the travel time from the delay puts it past the delay, then
            # do not process an additional delay, and simply go to the next step
            if delay >= 0:
                yield self.env.process(
                    self.weather_delay(delay, location="port", **shared_logging)
                )
            yield self.env.process(
                self.wait_until_next_shift(
                    **shared_logging,
                    **{"additional": "weather unsuitable to transfer crew"},
                )
            )
            yield self.env.process(
                self.travel(
                    start="port", end="site", set_current=system.id, **shared_logging
                )
            )
            yield self.env.process(
                self.crew_transfer(system, subassembly, request, to_system=to_system)
            )
            return

        yield self.env.process(
            self.weather_delay(delay, location="system", **shared_logging)
        )

        if to_system:
            additional = f"transferring crew from {self.settings.name} to {system.id}"
        else:
            additional = f"transferring crew from {system.id} to {self.settings.name}"
        self.env.log_action(
            action="transferring crew",
            additional=additional,
            duration=hours_to_process,
            salary_labor_cost=salary_cost,
            hourly_labor_cost=hourly_cost,
            equipment_cost=equipment_cost,
            location="system",
            **shared_logging,
        )
        self.transferring_crew = True
        yield self.env.timeout(hours_to_process)
        self.transferring_crew = False
        self.env.log_action(
            action="complete transfer",
            additional="complete",
            location="system",
            **shared_logging,
        )

    def mooring_connection(
        self,
        system: System,
        request: RepairRequest,
        which: str,
    ) -> Generator[Timeout | Process, None, None]:
        """The process of either umooring a floating turbine to prepare for towing it to
        port, or reconnecting it after its repairs have been completed.

        Parameters
        ----------
        system : System
            The System that needs unmooring/reconnecting.
        request : RepairRequest
            The repair to be processed.
        which : bool
            "unmoor" for unmooring the turbine, "reconnect" for reconnecting the
            turbine.

        Returns
        -------
        None
            None is returned when this is run recursively to not repeat the process.

        Yields
        ------
        Generator[Timeout | Process, None, None]
            Yields a timeout event for the unmooring/reconnection once an uninterrupted
            weather window can be found.
        """
        assert isinstance(self.settings, UnscheduledServiceEquipmentData)  # mypy check
        which = which.lower().strip()
        if which == "unmoor":
            hours_to_process = self.settings.unmoor_hours
        elif which == "reconnect":
            hours_to_process = self.settings.reconnection_hours
        else:
            raise ValueError(
                f"Only `unmoor` and `reconnect` are allowable inputs, not {which}"
            )

        salary_cost = self.calculate_salary_cost(hours_to_process)
        hourly_cost = self.calculate_hourly_cost(hours_to_process)
        equipment_cost = self.calculate_equipment_cost(hours_to_process)

        shared_logging = {
            "system_id": system.id,
            "system_name": system.name,
            "system_ol": system.operating_level,
            "agent": self.settings.name,
            "reason": request.details.description,
            "request_id": request.request_id,
        }

        which_text = "unmooring" if which == "unmoor" else "mooring reconnection"

        delay, shift_delay = self.find_uninterrupted_weather_window(hours_to_process)
        # If there is a shift delay, then wait try again.
        if shift_delay:
            yield self.env.process(
                self.weather_delay(delay, location="site", **shared_logging)
            )
            yield self.env.process(
                self.wait_until_next_shift(
                    **shared_logging,
                    **{
                        "location": "site",
                        "additional": f"weather unsuitable for {which_text}",
                    },
                )
            )
            yield self.env.process(
                self.mooring_connection(system, request, which=which)  # type: ignore
            )
            return

        # If no shift delay, then process any weather delays before dis/connection
        yield self.env.process(
            self.weather_delay(delay, location="site", **shared_logging)
        )

        if which == "unmoor":
            additional = f"Unmooring {system.id} to tow to port"
        else:
            additional = f"Reconnecting the mooring to {system.id}"
        self.env.log_action(
            action=which_text,
            additional=additional,
            duration=hours_to_process,
            salary_labor_cost=salary_cost,
            hourly_labor_cost=hourly_cost,
            equipment_cost=equipment_cost,
            location="system",
            **shared_logging,  # type: ignore
        )

        if which == "unmoor":
            yield self.env.timeout(self.settings.unmoor_hours)
        else:
            yield self.env.timeout(self.settings.reconnection_hours)

        self.env.log_action(
            action=f"complete {which_text}",
            additional="complete",
            **shared_logging,  # type: ignore
        )

    def in_situ_repair(
        self,
        request: RepairRequest,
        time_processed: int | float = 0,
        prior_operation_level: float = -1.0,
    ) -> Generator[Timeout | Process, None, None]:
        """Processes the repair including any weather and shift delays.

        Parameters
        ----------
        request : RepairRequest
            The ``Maintenance`` or ``Failure`` receiving attention.
        time_processed : int | float, optional
            Time that has already been processed, by default 0.
        prior_operation_level : float, optional
            The operating level of the ``System`` just before the repair has begun, by
            default -1.0.

        Yields
        ------
        Generator[Timeout | Process, None, None]
            Timeouts for the repair process.
        """
        """
        NOTE: THE PROCESS
        1. Travel to site/turbine, if not there already
        2. Transfer Crew
        3. Number of hours required, or hours until the end of the shift
        4. Do repairs for the above
        5. When shift/repair is complete, transfer crew
        6. Travel to next turbine, back to port

        7. Repeat until the repair is complete
        8. Exit this function
        """
        shift_delay = False

        if request.cable:
            system = self.windfarm.cable(request.system_id)
            cable = request.subassembly_id.split("::")[1:]
            subassembly = self.windfarm.graph.edges[cable]["cable"]
        else:
            system = self.windfarm.system(request.system_id)  # type: ignore
            subassembly = getattr(system, request.subassembly_id)

        starting_operational_level = max(
            prior_operation_level, subassembly.operating_level
        )
        shared_logging = {
            "system_id": system.id,
            "system_name": system.name,
            "part_id": subassembly.id,
            "part_name": subassembly.name,
            "system_ol": system.operating_level,
            "part_ol": subassembly.operating_level,
            "agent": self.settings.name,
            "reason": request.details.description,
            "request_id": request.request_id,
        }

        # Ensure there is enough time to transfer the crew back and forth with a buffer
        # of twice the travel time or travel back to port and try again the next shift
        start_shift = self.settings.workday_start
        end_shift = self.settings.workday_end
        current = self.env.simulation_time
        hours_required = request.details.time - time_processed
        hours_available = hours_until_future_hour(current, end_shift)

        if hours_available <= self.settings.crew_transfer_time * 4:
            yield self.env.process(
                self.travel(
                    start="site",
                    end="port",
                    **shared_logging,
                )
            )
            yield self.env.process(self.wait_until_next_shift(**shared_logging))

            yield self.env.process(
                self.in_situ_repair(
                    request, prior_operation_level=starting_operational_level
                )
            )
            return

        # Travel to site or the next system on site
        if not self.at_system and self.at_port:
            yield self.env.process(
                self.travel(
                    start="port", end="site", set_current=system.id, **shared_logging
                )
            )
            yield self.env.process(
                self.crew_transfer(system, subassembly, request, to_system=True)
            )
        elif self.at_system is not None and not self.at_port:
            yield self.env.process(
                self.travel(
                    start="system",
                    end="system",
                    set_current=system.id,
                    **shared_logging,
                )
            )
            yield self.env.process(
                self.crew_transfer(system, subassembly, request, to_system=True)
            )
        else:
            raise RuntimeError(f"{self.settings.name} is lost!")

        current = self.env.simulation_time

        # If the hours required is longer than the shift time, then reset the available
        # number of hours and the appropriate weather forecast, accounting for crew
        # transfer, otherwise get an adequate weather window, allowing for interruptions
        if not (start_shift == 0 and end_shift == 24):
            hours_available = hours_until_future_hour(current, end_shift)
            hours_available -= self.settings.crew_transfer_time
            *_, weather_forecast = self._weather_forecast(
                hours_available, which="repair"
            )
        else:
            _, weather_forecast, _ = self.find_interrupted_weather_window(
                hours_required
            )

        # Check that all times are within the windfarm's working hours and cut off any
        # time points that fall outside of the work shifts
        if hours_required > hours_available:
            shift_delay = True

        hours_processed = 0
        weather_delay_groups = consecutive_groups(np.where(~weather_forecast)[0])
        while weather_forecast.size > 0 and hours_available > 0:
            delays = np.where(~weather_forecast)[0]
            if delays.size > 0:
                hours_to_process = delays[0]
                delay = weather_delay_groups.pop(0).size
            else:
                hours_to_process = weather_forecast.size
                delay = 0
            weather_forecast = weather_forecast[hours_to_process + delay :]

            # If the delay is at the start, hours_to_process is 0, and a delay gets
            # processed, otherwise the crew works for the minimum of
            # ``hours_to_process`` or maximum time that can be worked until the shift's
            # end, and maxed out by the hours required for the actual repair.
            if hours_to_process > 0:
                if hours_available <= hours_to_process:
                    hours_to_process = hours_available
                else:
                    current = self.env.simulation_time
                    hours_to_process = hours_until_future_hour(
                        current, current.hour + int(hours_to_process)
                    )
                if hours_required < hours_to_process:
                    hours_to_process = hours_required
                # Ensure this gets the correct float hours to the start of the target
                # hour, unless the hours to process is between (0, 1]
                yield self.env.process(
                    self.process_repair(  # type: ignore
                        hours_to_process, request.details, **shared_logging
                    )
                )
                hours_processed += hours_to_process
                hours_available -= hours_to_process
                hours_required -= hours_to_process

            # If a delay is the first part of the process or a delay occurs after the
            # some work is performed, then that delay is processed here.
            if delay > 0 and hours_required > 0:
                current = self.env.simulation_time
                hours_to_process = hours_until_future_hour(
                    current, current.hour + delay
                )
                yield self.env.process(
                    self.weather_delay(
                        hours_to_process, location="system", **shared_logging
                    )
                )
                hours_available -= hours_to_process

        # Reached the end of the shift or the end of the repair, and need to unload crew
        # from the system
        yield self.env.process(
            self.crew_transfer(system, subassembly, request, to_system=False)
        )
        if shift_delay:
            yield self.env.process(
                self.travel(start="site", end="port", **shared_logging)
            )
            yield self.env.process(self.wait_until_next_shift(**shared_logging))

            yield self.env.process(
                self.in_situ_repair(
                    request,
                    time_processed=hours_processed + time_processed,
                    prior_operation_level=starting_operational_level,
                )
            )
            return

        # Register the repair
        self.register_repair_with_subassembly(
            subassembly, request, starting_operational_level
        )
        action = "maintenance" if isinstance(request.details, Maintenance) else "repair"
        self.env.log_action(
            system_id=system.id,
            system_name=system.name,
            part_id=subassembly.id,
            part_name=subassembly.name,
            system_ol=system.operating_level,
            part_ol=subassembly.operating_level,
            agent=self.settings.name,
            action=f"{action} complete",
            reason=request.details.description,
            materials_cost=request.details.materials,
            additional="complete",
            request_id=request.request_id,
            location="system",
        )

        # If this is the end of the shift, ensure that we're traveling back to port
        if not self.env.is_workshift(start_shift, end_shift):
            yield self.env.process(
                self.travel(start="site", end="port", **shared_logging)
            )

    def run_scheduled_in_situ(self) -> Generator[Process, None, None]:
        """Runs the simulation of in situ repairs for scheduled servicing equipment
        that have the `onsite` designation or don't require mobilization.

        Yields
        ------
        Generator[Process, None, None]
            The simulation.
        """
        assert isinstance(self.settings, ScheduledServiceEquipmentData)  # mypy controls

        # If the starting operation date is the same as the simulations, set to onsite
        if self.settings.operating_dates[0] == self.env.simulation_time.date():
            yield self.env.process(self.mobilize_scheduled())

        while True:
            # Wait for a valid operational period to start
            if self.env.simulation_time.date() not in self.settings.operating_dates_set:
                yield self.env.process(self.mobilize_scheduled())

            # Wait for next shift to start
            is_workshift = self.env.is_workshift(
                workday_start=self.settings.workday_start,
                workday_end=self.settings.workday_end,
            )
            if not is_workshift:
                yield self.env.process(
                    self.wait_until_next_shift(
                        **{
                            "agent": self.settings.name,
                            "reason": "not in working hours",
                            "additional": "not in working hours",
                        }
                    )
                )

            request = self.get_next_request()
            if request is None:
                if not self.at_port:
                    yield self.env.process(
                        self.travel(
                            start="site",
                            end="port",
                            reason="no requests",
                            agent=self.settings.name,
                        )
                    )
                yield self.env.process(
                    self.wait_until_next_shift(
                        **{
                            "agent": self.settings.name,
                            "reason": "no requests",
                            "additional": "no work requests, waiting until the next shift",  # noqa: disabl#501
                        }
                    )
                )
            else:
                request = request.value
                if request.cable:
                    system = self.windfarm.cable(request.system_id)
                else:
                    system = self.windfarm.system(request.system_id)  # type: ignore
                yield system.servicing
                self.manager.halt_requests_for_system(system)
                yield self.env.process(self.in_situ_repair(request))

    def run_unscheduled_in_situ(
        self, request: RepairRequest
    ) -> Generator[Process, None, None]:
        """Runs an in situ repair simulation for unscheduled servicing equipment, or
        those that have to be mobilized before performing repairs and maintenance.

        Parameters
        ----------
        request : RepairRequest
            The repair request that triggered the repair process.

        Yields
        ------
        Generator[Process, None, None]
            The simulation
        """
        self.dispatched = True
        assert isinstance(self.settings, UnscheduledServiceEquipmentData)  # mypy helper
        mobilization_days = self.settings.mobilization_days
        charter_end_env_time = self.settings.charter_days * HOURS_IN_DAY
        charter_end_env_time += mobilization_days * HOURS_IN_DAY
        charter_end_env_time += self.env.now  # type: ignore

        current = self.env.simulation_time
        charter_start = current + pd.Timedelta(mobilization_days, "D")
        charter_end = (
            current
            + pd.Timedelta(mobilization_days, "D")
            + pd.Timedelta(self.settings.charter_days, "D")
        )
        charter_period = set(pd.date_range(charter_start, charter_end).date)

        # If the charter period contains any non-operational date, then, try again
        # at the next available date after the end of the attempted charter period
        intersection = charter_period.intersection(
            self.settings.non_operational_dates_set
        )
        if intersection:
            hours_to_next = self.hours_to_next_operational_date(
                start_search_date=max(intersection),
                exclusion_days=mobilization_days,
            )
            self.env.log_action(
                agent=self.settings.name,
                action="delay",
                reason="non-operational period",
                additional="waiting for next operational period",
                duration=hours_to_next,
            )
            yield self.env.timeout(hours_to_next)  # type: ignore
            yield self.env.process(self.run_unscheduled_in_situ(request))
            return

        # Ensure the system isn't in service already during the checking stages, then
        # halt its ongoing processes for the current repair.
        # NOTE: port-based equipment will have already halted the system's processes.
        if not hasattr(self, "port"):
            if request.cable:
                system = self.windfarm.cable(request.system_id)
            else:
                system = self.windfarm.system(request.system_id)  # type: ignore
            yield system.servicing
            seconds_to_wait, *_ = (
                random_generator.integers(low=0, high=30, size=1) / 3600.0
            )
            yield self.env.timeout(seconds_to_wait)
            yield system.servicing
            self.manager.halt_requests_for_system(system)

        while True:
            if self.env.now >= charter_end_env_time:
                self.onsite = False
                self.at_port = False
                self.at_site = False
                self.at_system = False
                self.dispatched = False
                self.current_system = None
                self.env.log_action(
                    agent=self.settings.name,
                    action="leaving site",
                    reason="charter period has ended",
                )
                break

            if not self.onsite:
                yield self.env.process(self.mobilize())
                yield self.env.process(self.in_situ_repair(request))

            # Wait for next shift to start
            is_workshift = self.env.is_workshift(
                workday_start=self.settings.workday_start,
                workday_end=self.settings.workday_end,
            )
            if not is_workshift:
                yield self.env.process(
                    self.wait_until_next_shift(
                        **{
                            "agent": self.settings.name,
                            "reason": "not in working hours",
                            "additional": "not in working hours",
                        }
                    )
                )

            # Wait one second to ensure there are no timing collisions
            yield self.env.timeout(1.0 / 3600)
            request = self.get_next_request()
            if request is None:
                yield self.env.process(
                    self.wait_until_next_shift(
                        **{
                            "agent": self.settings.name,
                            "reason": "no requests",
                            "additional": "no work requests submitted by start of shift",  # noqa: disabl#501
                        }
                    )
                )
            else:
                request = request.value  # type: ignore
                if request.cable:
                    system = self.windfarm.cable(request.system_id)
                else:
                    system = self.windfarm.system(request.system_id)  # type: ignore
                yield system.servicing
                self.manager.halt_requests_for_system(system)
                yield self.env.process(self.in_situ_repair(request))
        self.dispatched = False

    def run_tow_to_port(self, request: RepairRequest) -> Generator[Process, None, None]:
        """Runs the tow to port logic, so a turbine can be repaired at port.

        Parameters
        ----------
        request : RepairRequest
            The request the triggered the tow-to-port strategy.

        Yields
        ------
        Generator[Process, None, None]
            The series of events that simulate the complete towing logic.

        Raises
        ------
        ValueError
            Raised if the equipment is not currently at port
        """
        self.dispatched = True
        system = self.windfarm.system(request.system_id)

        shared_logging = {
            "system_id": request.system_id,
            "system_name": request.system_name,
            "request_id": request.request_id,
            "agent": self.settings.name,
        }

        # Travel to the turbine
        yield self.env.process(
            self.travel(
                "port",
                "site",
                set_current=system.id,
                reason=request.details.description,
                **shared_logging,  # type: ignore
            )
        )

        # Unmoor the turbine and tow it back to port
        yield self.env.process(self.mooring_connection(system, request, which="unmoor"))
        yield self.env.process(
            self.tow("site", "port", reason="towing turbine to port", **shared_logging)
        )
        self.dispatched = False

    def run_tow_to_site(self, request: RepairRequest) -> Generator[Process, None, None]:
        """Runs the tow to site logic for after a turbine has had its repairs completed
        at port.

        Parameters
        ----------
        request : RepairRequest
            The request the triggered the tow-to-port strategy.

        Yields
        ------
        Generator[Process, None, None]
            The series of events that simulate the complete towing logic.

        Raises
        ------
        ValueError
            Raised if the equipment is not currently at port
        """
        self.dispatched = True
        system = self.windfarm.system(request.system_id)
        shared_logging = {
            "agent": self.settings.name,
            "request_id": request.request_id,
            "system_id": request.system_id,
            "system_name": request.system_name,
        }

        # Tow the turbine back to site and reconnect it to the mooring system
        yield self.env.process(
            self.tow(
                "port",
                "site",
                set_current=system.id,
                reason="towing turbine back to site",
                **shared_logging,
            )
        )
        yield self.env.process(
            self.mooring_connection(system, request, which="reconnect")  # type: ignore
        )

        # Reset the turbine back to operating and return to port
        reset_system_operations(system)
        self.manager.enable_requests_for_system(system)
        yield self.env.process(
            self.travel(
                "site",
                "port",
                reason="traveling back to port after returning turbine",
                **shared_logging,  # type: ignore
            )
        )
        self.dispatched = False



"""Creates the utilities objects."""


from .time import (
    HOURS_IN_DAY,
    HOURS_IN_YEAR,
    calculate_cost,
    check_working_hours,
    convert_dt_to_hours,
    hours_until_future_hour,
)
from .logging import setup_logger, format_events_log_message
from .utilities import IEC_power_curve, _mean, cache, create_variable_from_string



"""Tests wombat/core/repair_management.py."""


import numpy as np
import pytest

from wombat.core.library import load_yaml
from wombat.core.data_classes import (
    VALID_EQUIPMENT,
    StrategyMap,
    EquipmentMap,
    RepairRequest,
    ServiceEquipmentData,
)
from wombat.windfarm.windfarm import Windfarm
from wombat.core.repair_management import RepairManager
from wombat.core.service_equipment import ServiceEquipment
from wombat.windfarm.system.system import System
from wombat.windfarm.system.subassembly import Subassembly

from tests.conftest import VESTAS_V90, GENERATOR_SUBASSEMBLY


def test_repair_manager_init(env_setup):
    """Test the initialization steps."""
    env = env_setup

    # Test the capacity input
    manager = RepairManager(env, capacity=1)
    assert manager.env == env
    assert manager._current_id == 0
    assert manager.capacity == 1
    assert manager.items == []
    assert manager.downtime_based_equipment == StrategyMap()
    assert manager.request_based_equipment == StrategyMap()

    # Test the defaults
    manager = RepairManager(env)
    assert manager.env == env
    assert manager._current_id == 0
    assert manager.capacity == np.inf
    assert manager.items == []
    assert manager.downtime_based_equipment == StrategyMap()
    assert manager.request_based_equipment == StrategyMap()

    # Ensure the repair manager has an initialized windfarm registered
    windfarm = Windfarm(env, "layout.csv", manager)
    assert manager.windfarm == windfarm

    # Ensure the repair manager has an initialized equipment registered
    # NOTE: This will only check for the first addition, otherwise the StrategyMap tests
    # above, are sufficient

    # Add a downtime-based piece of equipment
    mapping = StrategyMap()
    fsv = load_yaml(env.data_dir / "vessels", "fsv_downtime.yaml")
    fsv_data = ServiceEquipmentData(fsv).determine_type()
    for capability in fsv_data.capability:
        mapping.update(capability, fsv_data.strategy_threshold, fsv_data)
    fsv = ServiceEquipment(env, windfarm, manager, "fsv_downtime.yaml")
    fsv_map = EquipmentMap(fsv.settings.strategy_threshold, fsv)
    assert manager.request_based_equipment == StrategyMap()
    assert manager.downtime_based_equipment.SCN == [fsv_map]
    assert manager.downtime_based_equipment.LCN == []
    assert manager.downtime_based_equipment.DRN == []
    assert manager.downtime_based_equipment.CTV == []
    assert manager.downtime_based_equipment.CAB == []
    assert manager.downtime_based_equipment.RMT == []
    assert manager.downtime_based_equipment.DSV == []

    # Add a requests-based piece of equipment
    fsv = ServiceEquipment(env, windfarm, manager, "fsv_requests.yaml")
    fsv_map = EquipmentMap(fsv.settings.strategy_threshold, fsv)
    assert manager.request_based_equipment.SCN == [fsv_map]
    assert manager.request_based_equipment.LCN == []
    assert manager.request_based_equipment.DRN == []
    assert manager.request_based_equipment.CTV == []
    assert manager.request_based_equipment.CAB == []
    assert manager.request_based_equipment.RMT == []
    assert manager.request_based_equipment.DSV == []

    # Check for registering of equipment with multiple capabilities
    hlv = ServiceEquipment(env, windfarm, manager, "hlv_requests_multi_capability.yaml")
    hlv_map = EquipmentMap(hlv.settings.strategy_threshold, hlv)
    assert manager.request_based_equipment.SCN == [fsv_map, hlv_map]
    assert manager.request_based_equipment.LCN == [hlv_map]
    assert manager.request_based_equipment.DRN == []
    assert manager.request_based_equipment.CTV == []
    assert manager.request_based_equipment.CAB == []
    assert manager.request_based_equipment.RMT == []
    assert manager.request_based_equipment.DSV == []

    # Ensure that a scheduled equipment type will raise an error, though this should
    # never be called within the code
    ctv = ServiceEquipment(env, windfarm, manager, "ctv_quick_load.yaml")
    with pytest.raises(ValueError):
        manager._register_equipment(ctv)


def test_register_request_and_submit_request_and_get_request(env_setup):
    """Tests the ``RepairManager.submit_request`` method."""
    # Set up all the required infrastructure
    env = env_setup
    manager = RepairManager(env)
    turbine = System(env, manager, "WTG001", "Vestas V90 001", VESTAS_V90, "turbine")
    generator = Subassembly(turbine, env, "GEN001", GENERATOR_SUBASSEMBLY)

    # Test that a submitted failure is registered correctly
    manual_reset = generator.data.failures[1]
    repair_request_failure = RepairRequest(
        turbine.id,
        turbine.name,
        generator.id,
        generator.name,
        manual_reset.level,
        manual_reset,
    )
    repair_request_failure = turbine.repair_manager.register_request(
        repair_request_failure
    )
    assert manager._current_id == 1
    assert repair_request_failure.request_id == "RPR00000000"

    turbine.repair_manager.submit_request(repair_request_failure)
    assert len(manager.items) == 1
    assert not manager.downtime_based_equipment.is_running
    assert not manager.request_based_equipment.is_running

    # Test that a submitted maintenance task is registered correctly
    annual_service = generator.data.maintenance[0]
    repair_request_maintenance = RepairRequest(
        turbine.id,
        turbine.name,
        generator.id,
        generator.name,
        annual_service.level,
        annual_service,
    )
    repair_request_maintenance = turbine.repair_manager.register_request(
        repair_request_maintenance
    )
    assert manager._current_id == 2
    assert repair_request_maintenance.request_id == "MNT00000001"

    turbine.repair_manager.submit_request(repair_request_maintenance)
    assert len(manager.items) == 2
    assert not manager.downtime_based_equipment.is_running
    assert not manager.request_based_equipment.is_running

    # Get the failure request back to ensure it's still the same object
    # NOTE: this will also purge it from the system
    retrieved_request = manager.get_request_by_system(["CTV"], system_id=turbine.id)
    assert retrieved_request.value == repair_request_failure

    # Ensure that the turbine has been labeled invalid so that another servicing
    # equipment can't access it
    assert manager.invalid_systems == [turbine.id]
    retrieved_request = manager.get_request_by_system(["CTV"], system_id=turbine.id)
    assert retrieved_request is None

    assert manager.items == [repair_request_maintenance]


def test_request_map(env_setup):
    """Tests the ``RepairManager.request_map`` property."""
    # Set up all the required infrastructure
    env = env_setup
    manager = RepairManager(env)
    turbine = System(env, manager, "WTG001", "Vestas V90 001", VESTAS_V90, "turbine")
    generator = Subassembly(turbine, env, "GEN001", GENERATOR_SUBASSEMBLY)

    # Test nothing is created yet
    assert manager.request_map == {}

    # Test that all the failures and maintenance tasks are registered correctly
    correct_request_map = {"CTV": 4, "SCN": 1, "LCN": 1}
    # Submit a request for all of the maintenance and failure objects in the generator
    for repair in generator.data.maintenance + [*generator.data.failures.values()]:
        repair_request = RepairRequest(
            turbine.id, turbine.name, generator.id, generator.name, repair.level, repair
        )
        repair_request = turbine.repair_manager.register_request(repair_request)
        turbine.repair_manager.submit_request(repair_request)

    assert manager.request_map == correct_request_map


def test_get_requests(env_setup):
    """Tests the ``RepairManager.get_request_by_system`` and
    ``RepairManager.get_request_by_severity`` methods.
    """
    # Set up all the required infrastructure
    env = env_setup
    manager = RepairManager(env)
    windfarm = Windfarm(env, "layout_single_subassembly.csv", manager)
    capability_list = VALID_EQUIPMENT
    turbine1 = windfarm.system("WTG001")
    turbine2 = windfarm.system("WTG002")
    turbine3 = windfarm.system("WTG003")
    generator1 = [el for el in turbine1.subassemblies if el.name == "generator"][0]
    generator2 = [el for el in turbine2.subassemblies if el.name == "generator"][0]
    generator3 = [el for el in turbine3.subassemblies if el.name == "generator"][0]

    # Ensure no requests are returned
    for _id in ("WTG001", "WTG002", "WTG003"):
        request = manager.get_request_by_system(capability_list, _id)
        assert request is None

    for severity in range(6):
        request = manager.get_request_by_severity(capability_list, severity)
        assert request is None

    # Submit a handful of requests with variables indicating what the request is and
    # which turbine it is originating from
    medium_repair1 = generator1.data.failures[3]
    medium_repair1 = RepairRequest(
        turbine1.id,
        turbine1.name,
        generator1.id,
        generator1.name,
        medium_repair1.level,
        medium_repair1,
    )
    medium_repair1 = turbine1.repair_manager.register_request(medium_repair1)
    turbine1.repair_manager.submit_request(medium_repair1)

    annual_reset2 = generator2.data.maintenance[0]
    annual_reset2 = RepairRequest(
        turbine2.id,
        turbine2.name,
        generator2.id,
        generator2.name,
        annual_reset2.level,
        annual_reset2,
    )
    annual_reset2 = turbine2.repair_manager.register_request(annual_reset2)
    turbine2.repair_manager.submit_request(annual_reset2)

    minor_repair2 = generator2.data.failures[2]
    minor_repair2 = RepairRequest(
        turbine2.id,
        turbine2.name,
        generator2.id,
        generator2.name,
        minor_repair2.level,
        minor_repair2,
    )
    minor_repair2 = turbine2.repair_manager.register_request(minor_repair2)
    turbine2.repair_manager.submit_request(minor_repair2)

    major_replacement3 = generator3.data.failures[5]
    major_replacement3 = RepairRequest(
        turbine3.id,
        turbine3.name,
        generator3.id,
        generator3.name,
        major_replacement3.level,
        major_replacement3,
    )
    major_replacement3 = turbine3.repair_manager.register_request(major_replacement3)
    turbine3.repair_manager.submit_request(major_replacement3)

    medium_repair3 = generator3.data.failures[3]
    medium_repair3 = RepairRequest(
        turbine3.id,
        turbine3.name,
        generator3.id,
        generator1.name,
        medium_repair3.level,
        medium_repair3,
    )
    medium_repair3 = turbine3.repair_manager.register_request(medium_repair3)
    turbine3.repair_manager.submit_request(medium_repair3)

    major_repair2 = generator2.data.failures[4]
    major_repair2 = RepairRequest(
        turbine2.id,
        turbine2.name,
        generator2.id,
        generator2.name,
        major_repair2.level,
        major_repair2,
    )
    major_repair2 = turbine2.repair_manager.register_request(major_repair2)
    turbine2.repair_manager.submit_request(major_repair2)

    # Basic checks that the submissions are all there
    assert len(manager.items) == 6

    # Test that the first submitted request is returned with no system filtering
    request = manager.get_request_by_system(capability_list)
    assert request.value == medium_repair1

    # Test that the major replacement is retrieved first when requesting by severity
    # level 5 and that other invalid requests return None
    request = manager.get_request_by_severity(capability_list, 6)
    assert request is None

    request = manager.get_request_by_severity(capability_list, 1)
    assert request is None

    request = manager.get_request_by_severity(capability_list, 5)
    assert request.value == major_replacement3

    # Ensure the next highest severity level request is (major repair 2) is picked
    request = manager.get_request_by_severity(capability_list)
    assert request.value == major_repair2

    # Test that the Turbine 2 requests are retrieved in order and with respect to
    # capability limitations and anything outside of the submitted capabilties is None
    manager.halt_requests_for_system(turbine2)
    request = manager.get_request_by_system(
        ["LCN", "CAB", "DSV", "RMT", "DRN"], turbine2.id
    )
    assert request is None
    manager.enable_requests_for_system(turbine2)

    request = manager.get_request_by_system(["CTV"], turbine2.id)
    manager.halt_requests_for_system(turbine2)
    assert request.value == annual_reset2
    manager.enable_requests_for_system(turbine2)

    request = manager.get_request_by_system(capability_list, turbine2.id)
    manager.halt_requests_for_system(turbine2)
    assert request.value == minor_repair2
    manager.enable_requests_for_system(turbine2)

    # With 1 requests left, ensure they're produced in the correct order and that
    # specifying incorrect severity levesls returns None
    request = manager.get_request_by_severity(capability_list, 4)
    assert request is None

    request = manager.get_request_by_severity(capability_list, 2)
    assert request is None

    # Activate turbine 3 and check that its medium repair is the next in the list
    manager.halt_requests_for_system(turbine3)
    manager.enable_requests_for_system(turbine3)
    request = manager.get_request_by_severity(capability_list, 3)
    assert request.value == medium_repair3


def test_purge_subassembly_requests(env_setup):
    """Tests the ``RepairManager.purge_subassembly_requests`` method."""
    env = env_setup
    manager = RepairManager(env)
    windfarm = Windfarm(env, "layout_with_gearboxes.csv", manager)
    turbine1 = windfarm.system("S00T1")
    turbine2 = windfarm.system("S00T2")
    turbine3 = windfarm.system("S00T3")
    generator1 = turbine1.generator
    gearbox1 = turbine1.gearbox
    generator2 = turbine2.generator
    gearbox2 = turbine2.gearbox
    generator3 = turbine3.generator
    gearbox3 = turbine3.gearbox

    # Check there are no submitted items
    assert manager.items == []

    # Check that None is returned when there are not items
    response = manager.purge_subassembly_requests(turbine1.id, generator1.id)
    assert response is None

    # Submit a handful of requests with variables indicating what the request is and
    # which turbine it is originating from
    gen_medium_repair1 = generator1.data.failures[3]
    gen_medium_repair1 = RepairRequest(
        turbine1.id,
        turbine1.name,
        generator1.id,
        generator1.name,
        gen_medium_repair1.level,
        gen_medium_repair1,
    )
    gen_medium_repair1 = turbine1.repair_manager.register_request(gen_medium_repair1)
    turbine1.repair_manager.submit_request(gen_medium_repair1)

    gen_major_repair1 = generator1.data.failures[4]
    gen_major_repair1 = RepairRequest(
        turbine1.id,
        turbine1.name,
        generator1.id,
        generator1.name,
        gen_major_repair1.level,
        gen_major_repair1,
    )
    gen_major_repair1 = turbine1.repair_manager.register_request(gen_major_repair1)
    turbine1.repair_manager.submit_request(gen_major_repair1)

    gbx_medium_repair1 = gearbox1.data.failures[3]
    gbx_medium_repair1 = RepairRequest(
        turbine1.id,
        turbine1.name,
        gearbox1.id,
        gearbox1.name,
        gbx_medium_repair1.level,
        gbx_medium_repair1,
    )
    gbx_medium_repair1 = turbine1.repair_manager.register_request(gbx_medium_repair1)
    turbine1.repair_manager.submit_request(gbx_medium_repair1)

    gen_medium_repair2 = generator2.data.failures[3]
    gen_medium_repair2 = RepairRequest(
        turbine2.id,
        turbine2.name,
        generator2.id,
        generator2.name,
        gen_medium_repair2.level,
        gen_medium_repair2,
    )
    gen_medium_repair2 = turbine2.repair_manager.register_request(gen_medium_repair2)
    turbine2.repair_manager.submit_request(gen_medium_repair2)

    gen_major_repair2 = generator2.data.failures[4]
    gen_major_repair2 = RepairRequest(
        turbine2.id,
        turbine2.name,
        generator2.id,
        generator2.name,
        gen_major_repair2.level,
        gen_major_repair2,
    )
    gen_major_repair2 = turbine2.repair_manager.register_request(gen_major_repair2)
    turbine2.repair_manager.submit_request(gen_major_repair2)

    gbx_medium_repair2 = gearbox2.data.failures[3]
    gbx_medium_repair2 = RepairRequest(
        turbine2.id,
        turbine2.name,
        gearbox2.id,
        gearbox2.name,
        gbx_medium_repair2.level,
        gbx_medium_repair2,
    )
    gbx_medium_repair2 = turbine2.repair_manager.register_request(gbx_medium_repair2)
    turbine2.repair_manager.submit_request(gbx_medium_repair2)

    gen_medium_repair3 = generator3.data.failures[3]
    gen_medium_repair3 = RepairRequest(
        turbine3.id,
        turbine3.name,
        generator3.id,
        generator3.name,
        gen_medium_repair3.level,
        gen_medium_repair3,
    )
    gen_medium_repair3 = turbine3.repair_manager.register_request(gen_medium_repair3)
    turbine3.repair_manager.submit_request(gen_medium_repair3)

    gen_major_repair3 = generator3.data.failures[4]
    gen_major_repair3 = RepairRequest(
        turbine3.id,
        turbine3.name,
        generator3.id,
        generator3.name,
        gen_major_repair3.level,
        gen_major_repair3,
    )
    gen_major_repair3 = turbine3.repair_manager.register_request(gen_major_repair3)
    turbine3.repair_manager.submit_request(gen_major_repair3)

    gbx_medium_repair3 = gearbox3.data.failures[3]
    gbx_medium_repair3 = RepairRequest(
        turbine3.id,
        turbine3.name,
        gearbox3.id,
        gearbox3.name,
        gbx_medium_repair3.level,
        gbx_medium_repair3,
    )
    gbx_medium_repair3 = turbine3.repair_manager.register_request(gbx_medium_repair3)
    turbine3.repair_manager.submit_request(gbx_medium_repair3)

    # Check for the correct number of submissions
    assert len(manager.items) == 9

    # Check that mismatched identifiers don't delete results
    for subassembly in turbine1.subassemblies:
        if subassembly.id in ("generator", "gearbox"):
            continue
        response = manager.purge_subassembly_requests(
            system_id=turbine1.id, subassembly_id=subassembly.id
        )
        assert response is None
        assert len(manager.items) == 9

    for subassembly in turbine2.subassemblies:
        if subassembly.id in ("generator", "gearbox"):
            continue
        response = manager.purge_subassembly_requests(
            system_id=turbine2.id, subassembly_id=subassembly.id
        )
        assert response is None
        assert len(manager.items) == 9

    for subassembly in turbine3.subassemblies:
        if subassembly.id in ("generator", "gearbox"):
            continue
        response = manager.purge_subassembly_requests(
            system_id=turbine3.id, subassembly_id=subassembly.id
        )
        assert response is None
        assert len(manager.items) == 9

    # Check that only the 1 gearbox submission for turbine 1 is all that is removed
    response = manager.purge_subassembly_requests(
        system_id=turbine1.id, subassembly_id=gearbox1.id
    )
    assert len(response) == 1
    assert response[0] == gbx_medium_repair1
    assert len(manager.items) == 8

    # Check that only the turbine 1 medium repair submission is removed
    response = manager.purge_subassembly_requests(
        system_id=turbine1.id,
        subassembly_id=generator1.id,
        exclude=[gen_major_repair1.request_id],
    )
    response = list(response)
    assert len(response) == 1
    assert response[0] == gen_medium_repair1
    assert len(manager.items) == 7

    # Check that only the turbine 1 major repair submission is removed
    response = manager.purge_subassembly_requests(
        system_id=turbine1.id, subassembly_id=generator1.id
    )
    response = list(response)
    assert len(response) == 1
    assert response[0] == gen_major_repair1
    assert len(manager.items) == 6

    # Check that the turbine 2 generator submissions are removed
    response = manager.purge_subassembly_requests(
        system_id=turbine2.id, subassembly_id=generator2.id
    )
    response = list(response)
    assert len(response) == 2
    assert response[0] == gen_medium_repair2
    assert response[1] == gen_major_repair2
    assert len(manager.items) == 4

    # Check that only the 1 gearbox submission for turbine 2 is all that is removed
    response = manager.purge_subassembly_requests(
        system_id=turbine2.id, subassembly_id=gearbox2.id
    )
    assert len(response) == 1
    assert response[0] == gbx_medium_repair2
    assert len(manager.items) == 3

    # Check that the turbine 3 generator submissions are removed
    response = manager.purge_subassembly_requests(
        system_id=turbine3.id, subassembly_id=generator3.id
    )
    response = list(response)
    assert len(response) == 2
    assert response[0] == gen_medium_repair3
    assert response[1] == gen_major_repair3
    assert len(manager.items) == 1

    # Check that only the 1 gearbox submission for turbine 3 is all that is removed
    response = manager.purge_subassembly_requests(
        system_id=turbine3.id, subassembly_id=gearbox3.id
    )
    assert len(response) == 1
    assert response[0] == gbx_medium_repair3
    assert len(manager.items) == 0



"""Creates the necessary repair classes."""


from __future__ import annotations

from typing import TYPE_CHECKING, Generator
from itertools import chain
from collections import Counter

import numpy as np
from simpy.resources.store import FilterStore, FilterStoreGet

from wombat.core import (
    Failure,
    Maintenance,
    StrategyMap,
    RepairRequest,
    WombatEnvironment,
    UnscheduledServiceEquipmentData,
)


if TYPE_CHECKING:
    from wombat.core import Port, ServiceEquipment
    from wombat.windfarm import Windfarm
    from wombat.windfarm.system import Cable, System


# Numpy random generation initialization
random_generator = np.random.default_rng(seed=42)


class RepairManager(FilterStore):
    """Provides a class to manage repair and maintenance tasks.

    Parameters
    ----------
    FilterStore : simpy.resources.store.FilterStore
        The ``simpy`` class on which RepairManager is based to manage the repair and
        maintenance tasks.
    env : wombat.core.WombatEnvironment
        The simulation environment.
    capacity : float
        The maximum number of tasks that can be submitted to the manager, by default
        ``np.inf``.

    Attributes
    ----------
    env : wombat.core.WombatEnvironment
        The simulation environment.
    windfarm: wombat.windfarm.Windfarm
        The simulated windfarm. This is only used for getting the operational capacity.
    _current_id : int
        The logged and auto-incrememented integer base for the ID generated for each
        submitted repair request.
    downtime_based_equipment: StrategyMap
        The mapping between downtime-based servicing equipment and their capabilities.
    request_based_equipment: StrategyMap
        The mapping between request-based servicing equipment and their capabilities.
    """

    def __init__(self, env: WombatEnvironment, capacity: float = np.inf) -> None:
        super().__init__(env, capacity)

        self.env = env
        self._current_id = 0
        self.invalid_systems: list[str] = []

        self.downtime_based_equipment = StrategyMap()
        self.request_based_equipment = StrategyMap()
        self.completed_requests = FilterStore(self.env)
        self.in_process_requests = FilterStore(self.env)

    def _update_equipment_map(self, service_equipment: ServiceEquipment) -> None:
        """Updates ``equipment_map`` with a provided servicing equipment object."""
        capability = service_equipment.settings.capability
        strategy = service_equipment.settings.strategy

        if strategy == "downtime":
            mapping = self.downtime_based_equipment
        elif strategy == "requests":
            mapping = self.request_based_equipment
        else:
            # Shouldn't be possible to get here!
            raise ValueError("Invalid servicing equipment!")

        assert isinstance(service_equipment.settings, UnscheduledServiceEquipmentData)
        strategy_threshold = service_equipment.settings.strategy_threshold
        if isinstance(capability, list):
            for c in capability:
                mapping.update(c, strategy_threshold, service_equipment)

    def _register_windfarm(self, windfarm: Windfarm) -> None:
        """Adds the simulation windfarm to the class attributes."""
        self.windfarm = windfarm

    def _register_equipment(self, service_equipment: ServiceEquipment) -> None:
        """Adds the servicing equipment to the class attributes and adds it to the
        capabilities mapping.
        """
        self._update_equipment_map(service_equipment)

    def _register_port(self, port: Port) -> None:
        """Registers the port with the repair manager, so that they can communicate as
        needed.

        Parameters
        ----------
        port : Port
            The port where repairs will occur.
        """
        self.port = port

    def _create_request_id(self, request: RepairRequest) -> str:
        """Creates a unique ``request_id`` to be logged in the ``request``.

        Parameters
        ----------
        request : RepairRequest
            The request object.

        Returns
        -------
        str
            An 11-digit identifier starting with "MNT" for maintenance tasks or "RPR"
            for repairs.

        Raises
        ------
        ValueError
            If the ``request.details`` property is not a ``Failure`` or ``Maintenance``
            object,
            then a ValueError will be raised.
        """
        if isinstance(request.details, Failure):
            prefix = "RPR"
        elif isinstance(request.details, Maintenance):
            prefix = "MNT"
        else:
            # Note this is a safety, and shouldn't be able to be reached
            raise ValueError("A valid ``RepairRequest`` must be submitted")

        request_id = f"{prefix}{str(self._current_id).zfill(8)}"
        self._current_id += 1
        return request_id

    def _run_equipment_downtime(self, request: RepairRequest) -> None | Generator:
        """Run any equipment that has a pending request where the current windfarm
        operating capacity is less than or equal to the servicing equipment's threshold.

        TODO: This methodology needs to better resolve dispatching every equipment
        relating to a request vs just the one(s) that are required. Basically, don't
        dispatch every available HLV, but just one plus one of every other capability
        category that has pending requests
        """
        # Add an initial check to help avoid simultaneous dispatching
        seconds_to_wait, *_ = random_generator.integers(low=0, high=10, size=1) / 3600.0
        yield self.env.timeout(seconds_to_wait)

        # Port-based servicing equipment should be handled by the port and does not
        # have an operating reduction threshold to meet at this time
        if "TOW" in request.details.service_equipment:
            if request.system_id not in self.port.invalid_systems:
                yield self.env.process(self.port.run_tow_to_port(request))
            return

        # Wait for the actual system or cable to be available
        if request.cable:
            yield self.windfarm.cable(request.system_id).servicing
        else:
            yield self.windfarm.system(request.system_id).servicing

        operating_capacity = self.windfarm.current_availability_wo_servicing
        for capability in self.request_map:
            equipment_mapping = self.downtime_based_equipment.get_mapping(capability)
            for i, equipment in enumerate(equipment_mapping):
                if operating_capacity > equipment.strategy_threshold:
                    continue

                # Avoid simultaneous dispatches by waiting a random number of seconds
                seconds_to_wait, *_ = (
                    random_generator.integers(low=0, high=30, size=1) / 3600.0
                )
                yield self.env.timeout(seconds_to_wait)

                equipment_obj = equipment.equipment
                if equipment_obj.dispatched:
                    continue

                # Equipment-based logic does not manage system availability, so
                # ensure it's available prior to dispatching, and double check in
                # case delays causing a timing collision
                if equipment_obj.port_based:
                    if request.system_id in self.port.invalid_systems:
                        break

                    # Dispatch-triggering request must be removed from the queue
                    _ = self.get(lambda x: x == request)

                    yield self.windfarm.system(request.system_id).servicing
                    self.env.process(self.port.run_unscheduled_in_situ(request))
                else:
                    # Dispatch-triggering request must be removed from the queue
                    _ = self.get(lambda x: x == request)

                    yield self.in_process_requests.put(request)
                    self.env.process(equipment_obj.run_unscheduled_in_situ(request))

                    # Move the dispatched capability to the end of list to ensure proper
                    # cycling of available servicing equipment
                    self.downtime_based_equipment.move_equipment_to_end(capability, i)

    def _run_equipment_requests(self, request: RepairRequest) -> None | Generator:
        """Run the first piece of equipment (if none are onsite) for each equipment
        capability category where the number of requests is greater than or equal to the
        equipment's threshold.
        """
        # Add an initial check to help avoid simultaneous dispatching
        seconds_to_wait, *_ = random_generator.integers(low=0, high=30, size=1) / 3600.0
        yield self.env.timeout(seconds_to_wait)

        # Port-based servicing equipment should be handled by the port and does not have
        # a requests-based threshold to meet at this time
        if "TOW" in request.details.service_equipment:
            if request.system_id not in self.port.invalid_systems:
                yield self.env.process(self.port.run_tow_to_port(request))
            return

        # Wait for the actual system or cable to be available
        if request.cable:
            yield self.windfarm.cable(request.system_id).servicing
        else:
            yield self.windfarm.system(request.system_id).servicing

        dispatched = None
        for capability, n_requests in self.request_map.items():
            # For a requests basis, the capability and submitted request must match
            if capability not in request.details.service_equipment:
                continue
            equipment_mapping = self.request_based_equipment.get_mapping(capability)
            for i, equipment in enumerate(equipment_mapping):
                if n_requests < equipment.strategy_threshold:
                    continue

                # Avoid simultaneous dispatches by waiting a random number of seconds
                seconds_to_wait, *_ = (
                    random_generator.integers(low=0, high=30, size=1) / 3600.0
                )
                yield self.env.timeout(seconds_to_wait)

                # Run only the first piece of equipment in the mapping list, but ensure
                # that it moves to the back of the line after being used
                equipment_obj = equipment.equipment
                if equipment_obj.dispatched:
                    break

                # Either run the repair logic from the port for port-based servicing
                # equipment, so that it can self-mangge or dispatch the servicing
                # equipment directly, when port is an implicitly modeled aspect
                if equipment_obj.port_based:
                    # Equipment-based logic does not manage system availability, so
                    # ensure it's available prior to dispatching
                    if request.system_id in self.port.invalid_systems:
                        break

                    # Dispatch-triggering request must be removed from the queue
                    _ = self.get(lambda x: x == request)

                    yield self.env.process(self.port.run_unscheduled_in_situ(request))
                else:
                    # Dispatch-triggering request must be removed from the queue
                    _ = self.get(lambda x: x == request)

                    yield self.in_process_requests.put(request)
                    yield self.env.process(
                        equipment_obj.run_unscheduled_in_situ(request)
                    )

                    # Move the dispatched capability to the end of list to ensure proper
                    # cycling of available servicing equipment
                    self.request_based_equipment.move_equipment_to_end(capability, i)
                dispatched = capability
                break

        # Double check the the number of reqeusts is still below the threshold following
        # the dispatching of a piece of servicing equipment. This mostly pertains to
        # highly frequent request with long repair times and low thresholds.
        if (
            dispatched is None
            or equipment_obj.port_based
            or dispatched not in self.request_map
        ):
            return
        n_requests = self.request_map[dispatched]
        threshold_check = [
            n_requests >= eq.strategy_threshold
            for eq in self.request_based_equipment.get_mapping(dispatched)
        ]
        if any(threshold_check):
            new_request_check = [
                x
                for x in self.items
                if dispatched in x.details.service_equipment
                and x not in self.completed_requests.items
                and x not in self.in_process_requests.items
            ]
            if new_request_check:
                new_request = self.get(lambda x: x == new_request_check[0]).value
                yield self.env.process(self._run_equipment_requests(new_request))

    def register_request(self, request: RepairRequest) -> RepairRequest:
        """The method to submit requests to the repair mananger and adds a unique
        identifier for logging.

        Parameters
        ----------
        request : RepairRequest
            The request that needs to be tracked and logged.

        Returns
        -------
        RepairRequest
            The same request as passed into the method, but with a unique identifier
            used for logging.
        """
        request_id = self._create_request_id(request)
        request.assign_id(request_id)
        self.put(request)
        return request

    def submit_request(self, request: RepairRequest) -> None:
        """The method to submit requests to the repair mananger and adds a unique
        identifier for logging.

        Parameters
        ----------
        request : RepairRequest
            The request that needs to be tracked and logged.

        Returns
        -------
        RepairRequest
            The same request as passed into the method, but with a unique identifier
            used for logging.
        """
        if self.downtime_based_equipment.is_running:
            self.env.process(self._run_equipment_downtime(request))
        if self.request_based_equipment.is_running:
            self.env.process(self._run_equipment_requests(request))

    def get_request_by_system(
        self, equipment_capability: list[str], system_id: str | None = None
    ) -> FilterStoreGet | None:
        """Gets all repair requests for a certain turbine with given a sequence of
        ``equipment_capability`` as long as it isn't registered as unable to be
        serviced.

        Parameters
        ----------
        equipment_capability : list[str]
            The capability of the servicing equipment requesting repairs to process.
        system_id : Optional[str], optional
            ID of the turbine or OSS; should correspond to ``System.id``, by default
            None. If None, then it will simply sort the list by ``System.id`` and return
            the first repair requested.

        Returns
        -------
        Optional[FilterStoreGet]
            The first repair request for the focal system or None, if self.items is
            empty, or there is no matching system.
        """
        if not self.items:
            return None

        if system_id in self.invalid_systems:
            return None

        equipment_capability = set(equipment_capability)  # type: ignore

        # Filter the requests by system
        requests = self.items
        if system_id is not None:
            requests = [el for el in self.items if el.system_id == system_id]
        if requests == []:
            return None

        # Filter the requests by equipment capability and return the first valid request
        assert isinstance(equipment_capability, set)
        for request in requests:
            if equipment_capability.intersection(request.details.service_equipment):
                # If this is the first request for the system, make sure no other
                # servicing equipment can access it
                if request.system_id not in self.invalid_systems:
                    self.invalid_systems.append(request.system_id)
                return self.get(lambda x: x == requests[0])

        # There were no matching equipment requirements to match the equipment
        # attempting to retrieve its next request
        return None

    def get_request_by_severity(
        self,
        equipment_capability: list[str] | set[str],
        severity_level: int | None = None,
    ) -> FilterStoreGet | None:
        """Gets the next repair request by ``severity_level``.

        Parameters
        ----------
        equipment_capability : list[str]
            The capability of the equipment requesting possible repairs to make.
        severity_level : int
            Severity level of focus, default None.

        Returns
        -------
        Optional[FilterStoreGet]
            Repair request meeting the required criteria.
        """
        if not self.items:
            return None

        equipment_capability = set(equipment_capability)

        requests = self.items
        if severity_level is not None:
            # Capture only the desired severity level, if specified
            requests = [
                el for el in self.items if (el.severity_level == severity_level)
            ]
            if requests == []:
                return None

        # Re-order requests by severity (high to low) and the order they were submitted
        # Need to ensure that the requests are in submission order in case any get put
        # back
        requests = sorted(requests, key=lambda x: x.severity_level, reverse=True)
        for request in requests:
            if request.cable:
                if not self.windfarm.cable(request.system_id).servicing.triggered:
                    continue
            else:
                if not self.windfarm.system(request.system_id).servicing.triggered:
                    continue
            if request.system_id not in self.invalid_systems:
                if equipment_capability.intersection(request.details.service_equipment):
                    self.invalid_systems.append(request.system_id)
                    return self.get(lambda x: x == request)

        # Ensure None is returned if nothing is found in the loop just as a FilterGet
        # would if allowed to oeprate without the above looping to identify multiple
        # criteria and acting on the request before it's processed
        return None

    def halt_requests_for_system(self, system: System | Cable) -> None:
        """Disables the ability for servicing equipment to service a specific system,
        sets the turbine status to be in servicing, and interrupts all the processes
        to turn off operations.

        Parameters
        ----------
        system_id : str
            The system to disable repairs.
        """
        if system.id not in self.invalid_systems:
            self.invalid_systems.append(system.id)
        if system.servicing.triggered:
            system.servicing = self.env.event()
            system.interrupt_all_subassembly_processes()
        else:
            raise RuntimeError(
                f"{self.env.simulation_time} {system.id} already being serviced"
            )

    def register_repair(self, repair: RepairRequest, port: bool = False) -> Generator:
        """Registers the repair as complete with the repair managiner.

        Parameters
        ----------
        repair : RepairRequest
            The repair that has been completed.
        port : bool, optional
            If True, indicates that a port handled the repair, otherwise that a managed
            servicing equipment handled the repair, by default False.

        Yields
        ------
        Generator
            The ``completed_requests.put()`` that registers completion.
        """
        if port:
            yield self.completed_requests.put(repair)
        else:
            request = yield self.in_process_requests.get(lambda x: x == repair)
            yield self.completed_requests.put(request)

    def enable_requests_for_system(self, system: System | Cable) -> None:
        """Reenables service equipment operations on the provided system.

        Parameters
        ----------
        system_id : System | Cable
            The ``System`` or ``Cable`` that can be operated on again.=
        """
        if system.servicing.triggered:
            raise RuntimeError(
                f"{self.env.simulation_time} Repairs were already completed"
                f" at {system.id}"
            )
        _ = self.invalid_systems.pop(self.invalid_systems.index(system.id))
        system.servicing.succeed()
        system.interrupt_all_subassembly_processes()

    def get_all_requests_for_system(
        self, agent: str, system_id: str
    ) -> list[RepairRequest] | None:
        """Gets all repair requests for a specific ``system_id``.

        Parameters
        ----------
        agent : str
            The name of the entity requesting all of a system's repair requests.
        system_id : Optional[str], optional
            ID of the turbine or OSS; should correspond to ``System.id``.
            the first repair requested.

        Returns
        -------
        Optional[list[RepairRequest]]
            All repair requests for a given system. If no matching requests are found,
            or there aren't any items in the queue yet, then None is returned.
        """
        if not self.items:
            return None

        # Filter the requests by system
        requests = self.items
        if system_id is not None:
            requests = [el for el in self.items if el.system_id == system_id]
        if requests == []:
            return None

        # Loop the requests and pop them from the queue
        for request in requests:
            self.env.log_action(
                system_id=request.system_id,
                system_name=request.system_name,
                part_id=request.subassembly_id,
                part_name=request.subassembly_name,
                system_ol=float("nan"),
                part_ol=float("nan"),
                agent=agent,
                action="requests being moved",
                reason="",
                request_id=request.request_id,
            )
            _ = self.get(lambda x: x == request)  # pylint: disable=W0640

        return requests

    def purge_subassembly_requests(
        self, system_id: str, subassembly_id: str, exclude: list[str] = []
    ) -> list[RepairRequest] | None:
        """Yields all the requests for a system/subassembly combination. This is
        intended to be used to remove erroneous requests after a subassembly has been
        replaced.

        Parameters
        ----------
        system_id : str
            Either the ``System.id`` or ``Cable.id``.
        subassembly_id : str
            Either the ``Subassembly.id`` or the ``Cable.id`` repeated for cables.
        exclude : list[str]
            A list of ``request_id`` to exclude from the purge. This is a specific use
            case for the combined cable system/subassembly, but can be to exclude
            certain requests from the purge.

        Yields
        ------
        Optional[list[RepairRequest]]
            All requests made to the repair manager for the provided system/subassembly
            combination. Returns None if self.items is empty or the loop terminates
            without finding what it is looking for.
        """
        if not self.items:
            return None

        requests = [
            request
            for request in self.items
            if (
                request.system_id == system_id
                and request.subassembly_id == subassembly_id
                and request.request_id not in exclude
            )
        ]
        if requests == []:
            return None

        for request in requests:
            self.env.log_action(
                system_id=request.system_id,
                system_name=request.system_name,
                part_id=request.subassembly_id,
                part_name=request.subassembly_name,
                system_ol=float("nan"),
                part_ol=float("nan"),
                agent="RepairManager",
                action="request canceled",
                reason="replacement required",
                request_id=request.request_id,
            )
            _ = self.get(lambda x: x == request)  # pylint: disable=W0640
        return requests

    @property
    def request_map(self) -> dict[str, int]:
        """Creates an updated mapping between the servicing equipment capabilities and
        the number of requests that fall into each capability category (nonzero values
        only).
        """
        requests = dict(
            Counter(
                chain.from_iterable(r.details.service_equipment for r in self.items)
            )
        )
        return requests
